"""
test_file_io.py
***************

Tests for src/utils/file_io.py
"""

import os
import glob
import sys
import tempfile
import shutil

try:
    import cPickle as pickle
except ImportError:
    import pickle

import numpy as np

from MTfit.utilities.unittest_utils import TestCase
from MTfit.utilities.file_io import csv2inv
from MTfit.utilities.file_io import parse_csv
from MTfit.utilities.file_io import _parse_csv_events
from MTfit.utilities.file_io import parse_hyp
from MTfit.utilities.file_io import full_pdf_output_dicts
from MTfit.utilities.file_io import hyp_output_dicts
from MTfit.utilities.file_io import hyp_output
from MTfit.utilities.file_io import read_binary_output
from MTfit.utilities.file_io import read_pickle_output
from MTfit.utilities.file_io import pickle_output
from MTfit.utilities.file_io import MATLAB_output
from MTfit.utilities.file_io import read_matlab_output
from MTfit.utilities.file_io import read_scatangle_output
from MTfit.utilities.file_io import unique_columns
from MTfit.utilities.file_io import convert_keys_to_unicode
from MTfit.utilities.file_io import convert_keys_from_unicode


class IOTestCase(TestCase):

    def setUp(self):
        self.cwd = os.getcwd()
        if sys.version_info >= (3, 0):
            self.tempdir = tempfile.TemporaryDirectory()
            os.chdir(self.tempdir.name)
        else:
            self.tempdir = tempfile.mkdtemp()
            os.chdir(self.tempdir)
        self.existing_csv_files = glob.glob('*.csv')
        self.existing_hyp_files = glob.glob('*.hyp')
        self.existing_out_files = glob.glob('*.out')
        self.existing_mat_files = glob.glob('*.mat')
        self.existing_scatangle_files = glob.glob('*.scatangle')

    def tearDown(self):
        for fname in glob.glob('*.csv'):
            if fname not in self.existing_csv_files:
                try:
                    os.remove(fname)
                except Exception:
                    print('Cannot remove ', fname)
        for fname in glob.glob('*.hyp'):
            if fname not in self.existing_hyp_files:
                try:
                    os.remove(fname)
                except Exception:
                    print('Cannot remove ', fname)
        for fname in glob.glob('*.out'):
            if fname not in self.existing_out_files:
                try:
                    os.remove(fname)
                except Exception:
                    print('Cannot remove ', fname)
        for fname in glob.glob('*.mat'):
            if fname not in self.existing_mat_files:
                try:
                    os.remove(fname)
                except Exception:
                    print('Cannot remove ', fname)
        for fname in glob.glob('*.scatangle'):
            if fname not in self.existing_scatangle_files:
                try:
                    os.remove(fname)
                except Exception:
                    print('Cannot remove ', fname)
        os.chdir(self.cwd)
        if sys.version_info >= (3, 0):
            self.tempdir.cleanup()
        else:
            try:
                shutil.rmtree(self.tempdir)
            except:
                pass

    def station_angles(self):
        out = "504.7\n"
        out += "S0271   231.1   154.7\n"
        out += "S0649   42.9    109.7\n"
        out += "S0484   21.2    145.4\n"
        out += "S0263   256.4   122.7\n"
        out += "S0142   197.4   137.6\n"
        out += "S0244   229.7   148.1\n"
        out += "S0415   75.6    122.8\n"
        out += "S0065   187.5   126.1\n"
        out += "S0362   85.3    128.2\n"
        out += "S0450   307.5   137.7\n"
        out += "S0534   355.8   138.2\n"
        out += "S0641   14.7    120.2\n"
        out += "S0155   123.5   117\n"
        out += "S0162   231.8   127.5\n"
        out += "S0650   45.9    108.2\n"
        out += "S0195   193.8   147.3\n"
        out += "S0517   53.7    124.2\n"
        out += "S0004   218.4   109.8\n"
        out += "S0588   12.9    128.6\n"
        out += "S0377   325.5   165.3\n"
        out += "S0618   29.4    120.5\n"
        out += "S0347   278.9   149.5\n"
        out += "S0529   326.1   131.7\n"
        out += "S0083   223.7   118.2\n"
        out += "S0595   42.6    117.8\n"
        out += "S0236   253.6   118.6\n"
        out += '\n'
        out += "504.7\n"
        out += "S0271   230.9   154.8\n"
        out += "S0649   43      109.8\n"
        out += "S0484   21.3    145.4\n"
        out += "S0263   256.4   122.8\n"
        out += "S0142   197.3   137.6\n"
        out += "S0244   229.6   148.1\n"
        out += "S0415   75.7    122.8\n"
        out += "S0065   187.4   126.1\n"
        out += "S0362   85.3    128.2\n"
        out += "S0450   307.5   137.8\n"
        out += "S0534   355.7   138.3\n"
        out += "S0641   14.8    120.3\n"
        out += "S0155   123.5   117.1\n"
        out += "S0162   231.7   127.6\n"
        out += "S0650   45.9    108.3\n"
        out += "S0195   193.6   147.3\n"
        out += "S0517   53.7    124.2\n"
        out += "S0004   218.3   109.9\n"
        out += "S0588   13      128.7\n"
        out += "S0377   325.7   165.4\n"
        out += "S0618   29.5    120.5\n"
        out += "S0347   278.9   149.6\n"
        out += "S0529   326.1   131.8\n"
        out += "S0083   223.7   118.2\n"
        out += "S0595   42.7    117.9\n"
        out += "S0236   253.5   118.7\n"
        out += "\n"
        return out

    def csv_file(self):
        csv = """UID=123,,,,
PPolarity,,,,
Name,Azimuth,TakeOffAngle,Measured,Error
S001,120,70,1,0.01
S002,160,60,-1,0.02
P/SHRMSAmplitudeRatio,,,,
Name,Azimuth,TakeOffAngle,Measured,Error
S003,110,10,1 2,0.05 0.04
S005,140,10,1 4,0.01 0.02
,,,,
PPolarity ,,,,
Name,Azimuth,TakeOffAngle,Measured,Error
S003,110,10,1,0.05"""
        open('csvtest.csv', 'w').write(csv)

    def hyp_file(self):
        hyp = """NLLOC "/home/djp61/NLLOC/NLLOC_TC2g/loc/obs.20150126.222324.grid0" "LOCATED" "Location completed."
SIGNATURE "   NLLoc:v6.00.0 03Feb2015 15h57m15"
COMMENT "Seiscat Location Run for Synthetics_1D_Grad_Single"
GRID  100 100 100  -5 -5 -0  0.1 0.1 0.1 PROB_DENSITY
SEARCH OCTREE nInitial 4000 nEvaluated 60000 smallestNodeSide 0.007734/0.007734/0.015625 oct_tree_integral 2.874797e-02 scatter_volume 2.867319e-02
HYPOCENTER  x 0.0312109 y 0.0234766 z 3.32031  OT 24.2751  ix -1 iy -1 iz -1
GEOGRAPHIC  OT 2015 01 26  22 23   24.2751  Lat 0.0399196 Long 0.0397218 Depth 3.32031
QUALITY  Pmax 2.85006e+141 MFmin 82.0143 MFmax 86.6127 RMS 0.00958958 Nphs 37 Gap 36.756 Dist 0.264995 Mamp  -9.9 0 Mdur  -9.9 0
VPVSRATIO  VpVsRatio 1.80732  Npair 15  Diff 0.4647
STATISTICS  ExpectX 0.0290815 Y 0.0222124 Z 3.356  CovXX 0.0084406 XY 0.000322566 XZ -0.00225914 YY 0.0102132 YZ -0.00403783 ZZ 0.046807 EllAz1  93.2974 Dip1  3.00778 Len1  0.171202 Az2  182.96 Dip2  -6.39314 Len2  0.185775 Len3  4.089675e-01
STAT_GEOG  ExpectLat 0.0399082 Long 0.0397027 Depth 3.356
TRANSFORM  LAMBERT RefEllipsoid Clarke-1880  LatOrig 0.039707  LongOrig 0.039441  FirstStdParal 0.003533  SecondStdParal 0.075882  RotCW 0.000000
QML_OriginQuality  assocPhCt 37  usedPhCt 37  assocStaCt -1  usedStaCt 22  depthPhCt -1  stdErr 0.00958958  azGap 36.756  secAzGap 58.4897  gtLevel -  minDist 0.264995 maxDist 5.47344 medDist 3.25213
QML_OriginUncertainty  horUnc -1  minHorUnc 0.138862  maxHorUnc 0.153692  azMaxHorUnc 9.99915
FOCALMECH  Hyp  0.0399196 0.0397218 3.32031 Mech  0 0 0 mf  0 nObs 0
PHASE ID Ins Cmp On Pha  FM Date     HrMn   Sec     Err  ErrMag    Coda      Amp       Per  >   TTpred    Res       Weight    StaLoc(X  Y         Z)        SDist    SAzim  RAz  RDip RQual    Tcorr 
S0325  ?    ?    ? P      U 20150126 2223        25 GAU      0.01        -1        -1        -1 >    0.7224    0.0004    1.2751   -0.1600   -0.1600    0.0000    0.2650 226.19 225.8 172.6  9     0.0000
S0273  ?    ?    ? P      U 20150126 2223     25.02 GAU      0.01        -1        -1        -1 >    0.7454   -0.0024    1.2649   -0.1600   -0.7999    0.0000    0.8453 193.07 196.4 161.8  9     0.0000
S0429  ?    ?    ? P      U 20150126 2223     25.04 GAU      0.02        -1        -1        -1 >    0.7623    0.0009    1.2266   -0.1600    1.1199    0.0000    1.1130 350.11 347.0 157.5  9     0.0000
S0374  ?    ?    ? P      U 20150126 2223     25.06 GAU      0.01        -1        -1        -1 >    0.7738    0.0096    1.2421   -1.1200    0.4800    0.0000    1.2384 291.63 295.7 154.8  9     0.0000
S0246  ?    ?    ? P      U 20150126 2223     25.05 GAU      0.02        -1        -1        -1 >    0.7742   -0.0009    1.2220   -0.4800   -1.1199    0.0000    1.2525 204.09 206.3 154.9  9     0.0000
S0246  ?    ?    ? S      ? 20150126 2223     25.64 GAU       0.1        -1        -1        -1 >    1.3626    0.0066    0.7080   -0.4800   -1.1199    0.0000    1.2525 204.09 206.3 154.9  9     0.0000
S0196  ?    ?    ? P      D 20150126 2223      25.1 GAU      0.02        -1        -1        -1 >    0.8191    0.0047    1.2020    0.1600   -1.7598    0.0000    1.7880 175.87 173.8 146.9  9     0.0000
S0196  ?    ?    ? S      ? 20150126 2223     25.74 GAU      0.02        -1        -1        -1 >    1.4416    0.0184    0.9519    0.1600   -1.7598    0.0000    1.7880 175.87 173.8 146.9  9     0.0000
S0198  ?    ?    ? P      U 20150126 2223     25.11 GAU      0.02        -1        -1        -1 >    0.8374   -0.0035    1.1959    0.8000   -1.7598    0.0000    1.9420 156.68 154.8 144.1  9     0.0000
S0198  ?    ?    ? S      ? 20150126 2223     25.77 GAU      0.02        -1        -1        -1 >    1.4739    0.0165    0.9446    0.8000   -1.7598    0.0000    1.9420 156.68 154.8 144.1  9     0.0000
S0384  ?    ?    ? P      U 20150126 2223     25.14 GAU      0.02        -1        -1        -1 >    0.8590    0.0052    1.1858    2.0800    0.4800    0.0000    2.0991  77.44  74.5 141.5  9     0.0000
S0384  ?    ?    ? S      ? 20150126 2223     25.81 GAU      0.02        -1        -1        -1 >    1.5118    0.0190    0.9286    2.0800    0.4800    0.0000    2.0991  77.44  74.5 141.5  9     0.0000
S0528  ?    ?    ? P      U 20150126 2223     25.24 GAU      0.02        -1        -1        -1 >    0.9667   -0.0015    1.1455   -1.7600    2.3998    0.0000    2.9758 322.99 323.2 130.7 10     0.0000
S0528  ?    ?    ? S      ? 20150126 2223     25.99 GAU      0.02        -1        -1        -1 >    1.7014    0.0113    0.8798   -1.7600    2.3998    0.0000    2.9758 322.99 323.2 130.7 10     0.0000
S0413  ?    ?    ? P      D 20150126 2223     25.26 GAU      0.02        -1        -1        -1 >    0.9854    0.0000    1.1383    3.0400    0.7999    0.0000    3.1074  75.53  74.1 128.3  9     0.0000
S0413  ?    ?    ? S      ? 20150126 2223     26.03 GAU      0.02        -1        -1        -1 >    1.7343    0.0188    0.8624    3.0400    0.7999    0.0000    3.1074  75.53  74.1 128.3  9     0.0000
S0527  ?    ?    ? P      U 20150126 2223     25.27 GAU      0.02        -1        -1        -1 >    0.9955   -0.0000    1.1343   -2.0800    2.3998    0.0000    3.1787 318.38 319.0 128.5 10     0.0000
S0527  ?    ?    ? S      ? 20150126 2223     26.04 GAU      0.02        -1        -1        -1 >    1.7521    0.0111    0.8650   -2.0800    2.3998    0.0000    3.1787 318.38 319.0 128.5 10     0.0000
S0465  ?    ?    ? P      U 20150126 2223     25.28 GAU      0.02        -1        -1        -1 >    1.0197   -0.0141    1.1080    3.0400    1.4399    0.0000    3.3255  64.79  64.3 125.8  9     0.0000
S0465  ?    ?    ? S      ? 20150126 2223     26.08 GAU      0.02        -1        -1        -1 >    1.7948    0.0088    0.8541    3.0400    1.4399    0.0000    3.3255  64.79  64.3 125.8  9     0.0000
S0516  ?    ?    ? P      U 20150126 2223      25.3 GAU      0.01        -1        -1        -1 >    1.0278   -0.0020    1.1450    2.7200    2.0798    0.0000    3.3850  52.59  52.3 125.8 10     0.0000
S0516  ?    ?    ? S      ? 20150126 2223     26.11 GAU      0.02        -1        -1        -1 >    1.8090    0.0249    0.8332    2.7200    2.0798    0.0000    3.3850  52.59  52.3 125.8 10     0.0000
S0579  ?    ?    ? P      U 20150126 2223     25.34 GAU      0.02        -1        -1        -1 >    1.0723   -0.0061    1.1011   -2.0800    3.0397    0.0000    3.6817 325.01 325.7 122.3 10     0.0000
S0579  ?    ?    ? S      ? 20150126 2223     26.16 GAU      0.02        -1        -1        -1 >    1.8873   -0.0029    0.8298   -2.0800    3.0397    0.0000    3.6817 325.01 325.7 122.3 10     0.0000
S0211  ?    ?    ? P      U 20150126 2223     25.33 GAU      0.02        -1        -1        -1 >    1.0709   -0.0148    1.0874   -3.3601   -1.4399    0.0000    3.6935 246.66 246.4 121.9  9     0.0000
S0211  ?    ?    ? S      ? 20150126 2223     26.16 GAU      0.02        -1        -1        -1 >    1.8848   -0.0004    0.8308   -3.3601   -1.4399    0.0000    3.6935 246.66 246.4 121.9  9     0.0000
S0207  ?    ?    ? P      D 20150126 2223     25.39 GAU      0.02        -1        -1        -1 >    1.1348   -0.0182    1.0558    3.6801   -1.7598    0.0000    4.0613 116.05 115.7 117.6  9     0.0000
S0207  ?    ?    ? S      ? 20150126 2223     26.28 GAU      0.02        -1        -1        -1 >    1.9974    0.0082    0.7992    3.6801   -1.7598    0.0000    4.0613 116.05 115.7 117.6  9     0.0000
S0287  ?    ?    ? S      ? 20150126 2223     26.27 GAU      0.02        -1        -1        -1 >    1.9840    0.0115    0.8010   -4.0000   -0.4800    0.0000    4.0626 262.88 261.7 117.6  9     0.0000
S0046  ?    ?    ? P      U 20150126 2223     25.43 GAU      0.02        -1        -1        -1 >    1.1596   -0.0025    1.0699    2.0800   -3.6796    0.0000    4.2321 151.05 150.8 116.3  9     0.0000
S0046  ?    ?    ? S      ? 20150126 2223     26.32 GAU      0.02        -1        -1        -1 >    2.0409    0.0051    0.7995    2.0800   -3.6796    0.0000    4.2321 151.05 150.8 116.3  9     0.0000
S0669  ?    ?    ? P      U 20150126 2223     25.44 GAU      0.02        -1        -1        -1 >    1.1753   -0.0081    1.0596    1.7600    3.9996    0.0000    4.3357  23.50  23.4 115.2  9     0.0000
S0208  ?    ?    ? P      U 20150126 2223     25.44 GAU      0.02        -1        -1        -1 >    1.1777   -0.0105    1.0556    4.0001   -1.7598    0.0000    4.3511 114.20 114.0 115.1  9     0.0000
S0208  ?    ?    ? S      ? 20150126 2223     26.35 GAU       0.1        -1        -1        -1 >    2.0729    0.0034    0.6289    4.0001   -1.7598    0.0000    4.3511 114.20 114.0 115.1  9     0.0000
S0675  ?    ?    ? P      U 20150126 2223     25.61 GAU      0.02        -1        -1        -1 >    1.3584   -0.0195    0.9755    3.6801    3.9996    0.0000    5.3967  42.54  42.2 107.0  9     0.0000
S0675  ?    ?    ? S      ? 20150126 2223     26.67 GAU      0.02        -1        -1        -1 >    2.3909    0.0086    0.7983    3.6801    3.9996    0.0000    5.3967  42.54  42.2 107.0  9     0.0000
S0002  ?    ?    ? S      ? 20150126 2223     26.66 GAU      0.02        -1        -1        -1 >    2.4029   -0.0135    0.7945   -3.6800   -3.9996    0.0000    5.4734 222.69 222.2 106.8  9     0.0000
END_PHASE
END_NLLOC
"""
        open('hyptest.hyp', 'w').write(hyp)

    def csv_events(self):
        csv = [['UID=123,,,,', 'PPolarity,,,,', 'Name,Azimuth,TakeOffAngle,Measured,Error', 'S001,120,70,1,0.01',
                'S002,160,60,-1,0.02', 'P/SHRMSAmplitudeRatio,,,,', 'Name,Azimuth,TakeOffAngle,Measured,Error',
                'S003,110,10,1 2,0.05 0.04', 'S005,140,10,1 4,0.01 0.02'], ['PPolarity ,,,,', 'Name,Azimuth,TakeOffAngle,Measured,Error', 'S003,110,10,1,0.05']]
        return csv

    def test_csv2inv(self):
        self.csv_file()
        self.assertTrue(os.path.exists('csvtest.csv'))
        csv2inv('csvtest.csv')
        self.assertTrue(os.path.exists('csvtest.inv'))
        d = pickle.load(open('csvtest.inv', 'rb'))
        self.assertEqual(len(d), 2)
        self.assertEqual(d[0]['UID'], '123')
        self.assertEqual(d[0]['PPolarity']['Stations']['Name'], ['S001', 'S002'])
        self.assertEqual(d[0]['PPolarity']['Measured'][0, 0], 1)
        self.assertEqual(d[0]['PPolarity']['Measured'][1, 0], -1)
        self.assertEqual(sorted(d[0].keys()), ['P/SHRMSAmplitudeRatio', 'PPolarity', 'UID'])
        self.assertEqual(d[0]['P/SHRMSAmplitudeRatio']['Error'][0, 0], 0.05)
        self.assertEqual(d[0]['P/SHRMSAmplitudeRatio']['Error'][0, 1], 0.04)
        self.assertEqual(d[0]['P/SHRMSAmplitudeRatio']['Error'][1, 0], 0.01)
        self.assertEqual(d[0]['P/SHRMSAmplitudeRatio']['Error'][1, 1], 0.02)
        self.assertEqual(d[1]['UID'], '2')
        os.remove('csvtest.csv')
        os.remove('csvtest.inv')

    def test_parse_csv(self):
        self.csv_file()
        d = parse_csv('csvtest.csv')
        self.assertEqual(len(d), 2)
        self.assertEqual(d[0]['UID'], '123')
        self.assertEqual(d[0]['PPolarity']['Stations']['Name'], ['S001', 'S002'])
        self.assertEqual(d[0]['PPolarity']['Measured'][0, 0], 1)
        self.assertEqual(d[0]['PPolarity']['Measured'][1, 0], -1)
        self.assertEqual(sorted(d[0].keys()), ['P/SHRMSAmplitudeRatio', 'PPolarity', 'UID'])
        self.assertEqual(d[0]['P/SHRMSAmplitudeRatio']['Error'][0, 0], 0.05)
        self.assertEqual(d[0]['P/SHRMSAmplitudeRatio']['Error'][0, 1], 0.04)
        self.assertEqual(d[0]['P/SHRMSAmplitudeRatio']['Error'][1, 0], 0.01)
        self.assertEqual(d[0]['P/SHRMSAmplitudeRatio']['Error'][1, 1], 0.02)
        self.assertEqual(d[1]['UID'], '2')

    def test__parse_csv_events(self):
        d = _parse_csv_events(self.csv_events())
        self.assertEqual(len(d), 2)
        self.assertEqual(d[0]['UID'], '123')
        self.assertEqual(d[0]['PPolarity']['Stations']['Name'], ['S001', 'S002'])
        self.assertEqual(d[0]['PPolarity']['Measured'][0, 0], 1)
        self.assertEqual(d[0]['PPolarity']['Measured'][1, 0], -1)
        self.assertEqual(sorted(d[0].keys()), ['P/SHRMSAmplitudeRatio', 'PPolarity', 'UID'])
        self.assertEqual(d[0]['P/SHRMSAmplitudeRatio']['Error'][0, 0], 0.05)
        self.assertEqual(d[0]['P/SHRMSAmplitudeRatio']['Error'][0, 1], 0.04)
        self.assertEqual(d[0]['P/SHRMSAmplitudeRatio']['Error'][1, 0], 0.01)
        self.assertEqual(d[0]['P/SHRMSAmplitudeRatio']['Error'][1, 1], 0.02)
        self.assertEqual(d[1]['UID'], '2')

    def test_parse_hyp(self):
        self.hyp_file()
        events = parse_hyp('hyptest.hyp')
        self.assertEqual(len(events), 1)
        self.assertTrue('PPolarity' in events[0].keys())
        self.assertTrue('UID' in events[0].keys())
        self.assertTrue('hyp_file' in events[0].keys())
        self.assertEqual(events[0]['UID'], "20150126222324275")
        self.assertEqual(len(events[0].keys()), 3)

    def test_full_pdf_output_dicts(self):
        self.hyp_file()
        events = parse_hyp('hyptest.hyp')
        event = events[0]
        mdict, sdict = full_pdf_output_dicts(event, False, {'moment_tensor_space': np.matrix([[1., 2.], [2., 1.], [1., 2.], [2., 1.], [
                                             1., 2.], [2., 1.]]), 'ln_pdf': np.matrix([0, 0.7]), 'dV': 1, 'probability': np.matrix([[1., 2.]]), 'total_number_samples': 400})
        self.assertTrue('Other' in mdict.keys())
        self.assertTrue('Events' in mdict.keys())
        self.assertTrue('Stations' in mdict.keys())
        self.assertTrue('MTSpace' in mdict['Events'].keys())
        self.assertTrue('Probability' in mdict['Events'].keys())
        self.assertTrue('ln_pdf' in mdict['Events'].keys())
        self.assertTrue('UID' in mdict['Events'].keys())
        self.assertTrue('NSamples' in mdict['Events'].keys())
        self.assertEqual(mdict['Stations'].shape, (20, 4))
        event['PPolarityProbability'] = event['PPolarity'].copy()
        event['PPolarityProbability']['Measured'] = np.matrix([[0.6, 0.4],
                                                               [0.7, 0.3],
                                                               [0.8, 0.2],
                                                               [0.67, 0.33],
                                                               [0.94, 0.06],
                                                               [0.32, 0.68],
                                                               [0.96, 0.04],
                                                               [0.76, 0.24],
                                                               [0.82, 0.18],
                                                               [0.12, 0.88],
                                                               [0.57, 0.43],
                                                               [0.68, 0.32],
                                                               [0.51, 0.49],
                                                               [0.68, 0.32],
                                                               [0.50, 0.50],
                                                               [0.02, 0.98],
                                                               [0.6, 0.4],
                                                               [0.7, 0.3],
                                                               [0.8, 0.2],
                                                               [0.67, 0.33]
                                                               ])
        event.pop('PPolarity')
        mdict, sdict = full_pdf_output_dicts(event, ['PPolarityProbability'], {'moment_tensor_space': np.matrix([[1., 2.], [2., 1.], [1., 2.], [
                                             2., 1.], [1., 2.], [2., 1.]]), 'ln_pdf': np.matrix([0, 0.7]), 'dV': 1, 'probability': np.matrix([[1., 2.]]), 'total_number_samples': 400})
        self.assertTrue('Other' in mdict.keys())
        self.assertTrue('Events' in mdict.keys())
        self.assertTrue('Stations' in mdict.keys())
        self.assertTrue('MTSpace' in mdict['Events'].keys())
        self.assertTrue('Probability' in mdict['Events'].keys())
        self.assertTrue('ln_pdf' in mdict['Events'].keys())
        self.assertTrue('UID' in mdict['Events'].keys())
        self.assertTrue('NSamples' in mdict['Events'].keys())
        self.assertEqual(mdict['Stations'].shape, (20, 4))
        self.assertTrue(all(mdict['Stations'][:, 3] == np.array(
            [0.6, 0.7, 0.8, 0.67, 0.94, -0.68, 0.96, 0.76, 0.82, -0.88, 0.57, 0.68, 0.51, 0.68, 0.5, -0.98, 0.6, 0.7, 0.8, 0.67])))
        mdict, sdict = full_pdf_output_dicts(event, False, {'moment_tensor_space': np.matrix([[1., 2.], [2., 1.], [1., 2.], [2., 1.], [
                                             1., 2.], [2., 1.]]), 'ln_pdf': np.matrix([0, 0.7]), 'dV': 1, 'probability': np.matrix([[1., 2.]]), 'total_number_samples': 400})
        self.assertTrue('Other' in mdict.keys())
        self.assertTrue('Events' in mdict.keys())
        self.assertTrue('Stations' in mdict.keys())
        self.assertTrue('MTSpace' in mdict['Events'].keys())
        self.assertTrue('Probability' in mdict['Events'].keys())
        self.assertTrue('ln_pdf' in mdict['Events'].keys())
        self.assertTrue('UID' in mdict['Events'].keys())
        self.assertTrue('NSamples' in mdict['Events'].keys())
        self.assertEqual(mdict['Stations'].shape, (20, 4))

    def test_hyp_output_dicts(self):
        self.hyp_file()
        events = parse_hyp('hyptest.hyp')
        event = events[0]
        a, b, c = hyp_output_dicts(event, False, {'moment_tensor_space': np.matrix([[1., 2.], [2., 1.], [1., 2.], [2., 1.], [
                                   1., 2.], [2., 1.]]), 'ln_pdf': np.matrix([0, 0.7]), 'probability': np.matrix([[1., 2.]]), 'total_number_samples': 400})
        self.assertEqual(len(a.split('\n')), len(event['hyp_file'])+1)
        self.assertTrue('MOMENTTENSOR' in a)
        self.assertEqual(float(a.split()[a.split().index('MTNN')+1]), 2.0)
        self.assertEqual(len(b), 169)
        events = parse_hyp('hyptest.hyp')
        event = events[0]
        a, b, c = hyp_output_dicts(event, False, {'moment_tensor_space': np.matrix([[1., 2., 3.], [2., 1., 3.], [1., 2., 3.], [2., 1., 3.], [
                                   1., 2., 3.], [2., 1., 3.]]), 'ln_pdf': np.matrix([0, 0.7, 0]), 'probability': np.matrix([[1., 2., 1.]]), 'total_number_samples': 400})
        self.assertEqual(len(a.split('\n')), len(event['hyp_file'])+1)
        self.assertTrue('MOMENTTENSOR' in a)
        self.assertEqual(float(a.split()[a.split().index('MTNN')+1]), 2.0)
        self.assertEqual(len(b), 233)
        events = parse_hyp('hyptest.hyp')
        event = events[0]
        event['hyp_file'].pop(14)
        a, b, c = hyp_output_dicts(event, False, {'moment_tensor_space': np.matrix([[1., 2., 3.], [2., 1., 3.], [1., 2., 3.], [2., 1., 3.], [
                                   1., 2., 3.], [2., 1., 3.]]), 'ln_pdf': np.matrix([0, 0.7, 0]), 'probability': np.matrix([[1., 2., 1.]]), 'total_number_samples': 400})
        self.assertEqual(len(a.split('\n')), len(event['hyp_file'])+2)
        self.assertTrue('MOMENTTENSOR' in a)
        self.assertEqual(float(a.split()[a.split().index('MTNN')+1]), 2.0)
        self.assertEqual(len(b), 233)
        event.pop('hyp_file')
        a, b, c = hyp_output_dicts(event, False, {'moment_tensor_space': np.matrix([[1., 2., 3.], [2., 1., 3.], [1., 2., 3.], [2., 1., 3.], [
                                   1., 2., 3.], [2., 1., 3.]]), 'ln_pdf': np.matrix([0, 0.7, 0]), 'probability': np.matrix([[1., 2., 1.]]), 'total_number_samples': 400})
        self.assertEqual(len(a.split('\n')), 36)
        self.assertTrue('MOMENTTENSOR' in a)
        self.assertEqual(float(a.split()[a.split().index('MTNN')+1]), 2.0)
        self.assertEqual(len(b), 233)
        events = parse_hyp('hyptest.hyp')
        event = events[0]
        event['hyp_file'].pop(14)
        a, b, c = hyp_output_dicts(event, False, {'moment_tensor_space': np.matrix([[1., -0.51969334, 3.], [2., 0.22610635, 3.], [1., 0.29358698, 3.], [2., 0.58532165, 3.], [
                                   1., -0.27015115, 3.], [2., -0.42073549, 3.]]), 'ln_pdf': np.matrix([0, 0.7, 0]), 'probability': np.matrix([[1., 2., 1.]]), 'total_number_samples': 400})
        self.assertEqual(len(a.split('\n')), len(event['hyp_file'])+2)
        self.assertFalse('MOMENTTENSOR' in a)
        self.assertTrue('FOCALMECH' in a)
        self.assertAlmostEqual(
            float(a.split()[a.split().index('Mech')+1]), 0.5*180/np.pi, 5)
        self.assertEqual(len(b), 233)
        a, b, c = hyp_output_dicts(event, False, {'moment_tensor_space': np.matrix([[1., -0.51969334, 3.], [2., 0.22610635, 3.], [1., 0.29358698, 3.], [2., 0.58532165, 3.], [1., -0.27015115, 3.], [2., -0.42073549, 3.]]),
                                                  'probability': np.matrix([[1., 2., 1.]]), 'ln_pdf': np.matrix([0, 0.7, 0]), 'total_number_samples': 400, 'g': np.array([0.2, 0.2, 0.2]), 'd': np.array([0.2, 0.2, 0.2]), 'k': np.array([0.2, 0.2, 0.2]),
                                                  'h': np.array([0.2, 0.2, 0.2]), 's': np.array([0.2, 0.2, 0.2]), 'S1': np.array([0.2, 0.2, 0.2]), 'D1': np.array([0.2, 0.2, 0.2]), 'R1': np.array([0.2, 0.2, 0.2]), 'u': np.array([0.2, 0.2, 0.2]),
                                                  'v': np.array([0.2, 0.2, 0.2]), 'S2': np.array([0.2, 0.2, 0.2]), 'D2': np.array([0.2, 0.2, 0.2]), 'R2': np.array([0.2, 0.2, 0.2]), 'ln_bayesian_evidence': (1, 10)})
        self.assertEqual(len(a.split('\n')), len(event['hyp_file'])+2)
        self.assertFalse('MOMENTTENSOR' in a)
        self.assertTrue('FOCALMECH' in a)
        self.assertAlmostEqual(
            float(a.split()[a.split().index('Mech')+1]), 0.5*180/np.pi, 5)
        self.assertEqual(len(b), 545)

    def test_hyp_output(self):
        self.hyp_file()
        events = parse_hyp('hyptest.hyp')
        event = events[0]
        a, b, c = hyp_output_dicts(event, False, {'moment_tensor_space': np.matrix([[1., 2.], [2., 1.], [1., 2.], [2., 1.], [
                                   1., 2.], [2., 1.]]), 'ln_pdf': np.matrix([0, 0.7, 0]), 'probability': np.matrix([[1., 2.]]), 'total_number_samples': 400})
        try:
            os.remove('MTfitOUTPUTTEST.hyp')
        except Exception:
            pass
        try:
            os.remove('MTfitOUTPUTTEST.mt')
        except Exception:
            pass
        self.assertFalse(os.path.exists('MTfitOUTPUTTEST.hyp'))
        self.assertFalse(os.path.exists('MTfitOUTPUTTEST.mt'))
        fid, out_str = hyp_output([a, b], fid='MTfitOUTPUTTEST.hyp')
        self.assertTrue(os.path.exists('MTfitOUTPUTTEST.hyp'))
        self.assertTrue(os.path.exists('MTfitOUTPUTTEST.mt'))
        try:
            os.remove('MTfitOUTPUTTEST.hyp')
        except Exception:
            pass
        try:
            os.remove('MTfitOUTPUTTEST.mt')
        except Exception:
            pass
        event['hyp_file'].pop(14)
        a, b, c = hyp_output_dicts(event, False, {'moment_tensor_space': np.matrix([[1., -0.51969334, 3.], [2., 0.22610635, 3.], [1., 0.29358698, 3.], [2., 0.58532165, 3.], [1., -0.27015115, 3.], [2., -0.42073549, 3.]]),
                                                  'probability': np.matrix([[1., 2., 1.]]), 'ln_pdf': np.matrix([0, 0.7, 0]), 'total_number_samples': 400, 'g': np.array([0.2, 0.2, 0.2]), 'd': np.array([0.2, 0.2, 0.2]), 'k': np.array([0.2, 0.2, 0.2]),
                                                  'h': np.array([0.2, 0.2, 0.2]), 's': np.array([0.2, 0.2, 0.2]), 'S1': np.array([0.2, 0.2, 0.2]), 'D1': np.array([0.2, 0.2, 0.2]), 'R1': np.array([0.2, 0.2, 0.2]), 'u': np.array([0.2, 0.2, 0.2]),
                                                  'v': np.array([0.2, 0.2, 0.2]), 'S2': np.array([0.2, 0.2, 0.2]), 'D2': np.array([0.2, 0.2, 0.2]), 'R2': np.array([0.2, 0.2, 0.2]), 'ln_bayesian_evidence': 1.+10})
        try:
            os.remove('MTfitOUTPUTTEST.hyp')
        except Exception:
            pass
        try:
            os.remove('MTfitOUTPUTTEST.mt')
        except Exception:
            pass
        self.assertFalse(os.path.exists('MTfitOUTPUTTEST.hyp'))
        self.assertFalse(os.path.exists('MTfitOUTPUTTEST.mt'))
        fid, out_str = hyp_output([a, b], fid='MTfitOUTPUTTEST.hyp')
        self.assertTrue(os.path.exists('MTfitOUTPUTTEST.hyp'))
        self.assertTrue(os.path.exists('MTfitOUTPUTTEST.mt'))
        try:
            os.remove('MTfitOUTPUTTEST.hyp')
        except Exception:
            pass
        try:
            os.remove('MTfitOUTPUTTEST.mt')
        except Exception:
            pass

    def test_read_binary_output(self):
        self.hyp_file()
        events = parse_hyp('hyptest.hyp')
        event = events[0]
        x = {'moment_tensor_space': np.matrix([[1., 2.], [2., 1.], [1., 2.], [2., 1.], [1., 2.], [2., 1.]]), 'dkl': 1.2, 'ln_pdf': np.matrix(
            [0, 0.7]), 'probability': np.matrix([[1., 2.]]), 'total_number_samples': 400}
        a, b, c = hyp_output_dicts(event, False, x)
        try:
            os.remove('MTfitOUTPUTTEST.hyp')
        except Exception:
            pass
        try:
            os.remove('MTfitOUTPUTTEST.mt')
        except Exception:
            pass
        self.assertFalse(os.path.exists('MTfitOUTPUTTEST.hyp'))
        self.assertFalse(os.path.exists('MTfitOUTPUTTEST.mt'))
        fid, out_str = hyp_output([a, b], fid='MTfitOUTPUTTEST.hyp')
        y = read_binary_output('MTfitOUTPUTTEST.mt')
        self.assertTrue(os.path.exists('MTfitOUTPUTTEST.hyp'))
        self.assertTrue(os.path.exists('MTfitOUTPUTTEST.mt'))
        self.assertEqual(sorted(y[0].keys()), sorted(x.keys()))
        self.assertTrue(
            (y[0]['moment_tensor_space'] == x['moment_tensor_space']).all())
        self.assertEqual(y[0]['dkl'], x['dkl'])
        try:
            os.remove('MTfitOUTPUTTEST.hyp')
        except Exception:
            pass
        try:
            os.remove('MTfitOUTPUTTEST.mt')
        except Exception:
            pass
        x = {'moment_tensor_space': np.matrix([[1., -0.51969334, 3.], [2., 0.22610635, 3.], [1., 0.29358698, 3.], [2., 0.58532165, 3.], [1., -0.27015115, 3.], [2., -0.42073549, 3.]]),
             'probability': np.matrix([[1., 2., 1.]]), 'dkl': 2.4, 'ln_pdf': np.matrix([0, 0.7, 0]), 'total_number_samples': 400, 'g': np.array([0.1, 0.2, 0.3]), 'd': np.array([0.2, 0.2, 0.2]), 'k': np.array([0.2, 0.2, 0.2]),
             'h': np.array([0.2, 0.2, 0.2]), 's': np.array([0.2, 0.2, 0.2]), 'S1': np.array([0.2, 0.2, 0.2]), 'D1': np.array([0.2, 0.2, 0.2]), 'R1': np.array([0.2, 0.2, 0.2]), 'u': np.array([0.2, 0.2, 0.2]),
             'v': np.array([0.2, 0.2, 0.2]), 'S2': np.array([0.2, 0.2, 0.2]), 'D2': np.array([0.2, 0.2, 0.2]), 'R2': np.array([0.2, 0.2, 0.2]), 'ln_bayesian_evidence': 1.+10.}
        a, b, c = hyp_output_dicts(event, False, x)
        try:
            os.remove('MTfitOUTPUTTEST.hyp')
        except Exception:
            pass
        try:
            os.remove('MTfitOUTPUTTEST.mt')
        except Exception:
            pass
        self.assertFalse(os.path.exists('MTfitOUTPUTTEST.hyp'))
        self.assertFalse(os.path.exists('MTfitOUTPUTTEST.mt'))
        fid, out_str = hyp_output([a, b], fid='MTfitOUTPUTTEST.hyp')
        y = read_binary_output('MTfitOUTPUTTEST.mt')
        self.assertTrue(os.path.exists('MTfitOUTPUTTEST.hyp'))
        self.assertTrue(os.path.exists('MTfitOUTPUTTEST.mt'))
        self.assertEqual(sorted(y[0].keys()), sorted(x.keys()))
        self.assertTrue((y[0]['g'] == x['g']).all())
        self.assertEqual(
            y[0]['ln_bayesian_evidence'], x['ln_bayesian_evidence'])
        self.assertEqual(y[0]['dkl'], x['dkl'])
        try:
            os.remove('MTfitOUTPUTTEST.hyp')
        except Exception:
            pass
        try:
            os.remove('MTfitOUTPUTTEST.mt')
        except Exception:
            pass

    def test_read_pickle_output(self):
        self.hyp_file()
        events = parse_hyp('hyptest.hyp')
        event = events[0]
        mdict, sdict = full_pdf_output_dicts(event, False, {'moment_tensor_space': np.matrix([[1., 2.], [2., 1.], [1., 2.], [2., 1.], [
                                             1., 2.], [2., 1.]]), 'ln_pdf': np.matrix([0, 0.7]), 'dV': 1, 'probability': np.matrix([[1., 2.]]), 'total_number_samples': 400})
        pickle_output([mdict, sdict], 'hyptest.out')
        event, stations = read_pickle_output('hyptest.out')
        self.assertAlmostEqual(event, mdict['Events'], 10)
        self.assertAlmostEqual(
            stations['azimuth'], mdict['Stations'][:, 1], 10)
        self.assertAlmostEqual(
            stations['takeoff_angle'], mdict['Stations'][:, 2], 10)
        self.assertAlmostEqual(
            stations['polarity'], mdict['Stations'][:, 3], 10)

    def test_read_matlab_output(self):
        self.hyp_file()
        events = parse_hyp('hyptest.hyp')
        event = events[0]
        mdict, sdict = full_pdf_output_dicts(event, False, {'moment_tensor_space': np.matrix([[1., 2.], [2., 1.], [1., 2.], [2., 1.], [
                                             1., 2.], [2., 1.]]), 'ln_pdf': np.matrix([0, 0.7]), 'dV': 1, 'probability': np.matrix([[1., 2.]]), 'total_number_samples': 400})
        MATLAB_output([mdict, sdict], 'hyptest.mat')
        event, stations = read_matlab_output('hyptest.mat')
        self.assertAlmostEqual(event, mdict['Events'], 10)
        self.assertAlmostEqual(
            stations['azimuth'], mdict['Stations'][:, 1], 10)
        self.assertAlmostEqual(
            stations['takeoff_angle'], mdict['Stations'][:, 2], 10)
        self.assertAlmostEqual(
            stations['polarity'], mdict['Stations'][:, 3], 10)

    def test_read_scatangle_output(self):
        open('test_scatangle_file.scatangle', 'w').write(self.station_angles())
        data = read_scatangle_output('test_scatangle_file.scatangle')
        self.assertTrue('distribution' in data.keys())
        self.assertTrue('probability' in data.keys())
        self.assertEqual(len(data['probability']), 2)
        self.assertEqual(len(data['distribution']), 2)
        self.assertTrue('takeoff_angle' in data['distribution'][0].keys())
        self.assertTrue('azimuth' in data['distribution'][0].keys())
        self.assertTrue('names' in data['distribution'][0].keys())


class UtilsTestCase(TestCase):

    def test_unique_columns(self):
        data = np.matrix([[1, 0, 0, 2, 0, 1, 0],
                          [0, 0, 0, -1, 0, 0, 0],
                          [-1, 0, 0, -1, 0, -1, 0],
                          [0, 0, 1, 0, 1, 0, 1],
                          [0, 1, 0, 0, 0, 0, 0],
                          [0, 0, 0, 0, 0, 0, 0]])
        unique = unique_columns(data)
        self.assertEqual(unique.shape, (6, 4))
        unique, counts = unique_columns(data, counts=True)
        self.assertEqual(unique.shape, (6, 4))
        self.assertEqual(set(counts), set([1, 2, 3]))
        unique, counts, index = unique_columns(data, counts=True, index=True)
        self.assertEqual(unique.shape, (6, 4))
        self.assertEqual(set(counts), set([1, 2, 3]))
        self.assertEqual(set(index), set([3, 0, 2, 1]))
        unique, index = unique_columns(data, counts=False, index=True)
        self.assertEqual(unique.shape, (6, 4))
        self.assertEqual(set(index), set([3, 0, 2, 1]))

    def test_convert_keys_to_unicode(self):
        test = {'a': 1, 'b': {'c': 2}}
        result = convert_keys_to_unicode(test)
        if sys.version_info.major > 2:
            self.assertTrue(all([isinstance(u, str) for u in result.keys()]))
            self.assertTrue(all([isinstance(u, str) for u in result['b'].keys()]))
        else:
            self.assertTrue(all([isinstance(u, unicode) for u in result.keys()]))
            self.assertTrue(all([isinstance(u, unicode) for u in result['b'].keys()]))

    def test_convert_keys_from_unicode(self):
        test = {'a': 1, 'b': {'c': 2}}
        result = convert_keys_to_unicode(test)
        if sys.version_info.major > 2:
            self.assertTrue(all([isinstance(u, str) for u in result.keys()]))
            self.assertTrue(all([isinstance(u, str) for u in result['b'].keys()]))
        else:
            self.assertTrue(all([isinstance(u, unicode) for u in result.keys()]))
            self.assertTrue(all([isinstance(u, unicode) for u in result['b'].keys()]))
        result = convert_keys_from_unicode(result)
        self.assertTrue(all([isinstance(u, str) for u in result.keys()]))
        self.assertTrue(all([isinstance(u, str) for u in result['b'].keys()]))



import unittest
from types import MethodType
import sys

import numpy as np

from MTfit.utilities import C_EXTENSION_FALLBACK_LOG_MSG
from MTfit.utilities.unittest_utils import get_extension_skip_if_args
import MTfit.algorithms.base as base

if sys.version_info >= (3, 3):
    from unittest import mock
else:
    import mock


C_EXTENSIONS = get_extension_skip_if_args('MTfit.probability.cprobability')


class PythonOnly(object):

    def __enter__(self, *args, **kwargs):
        self.cprobability = base.cprobability
        base.cprobability = False

    def __exit__(self, *args, **kwargs):
        base.cprobability = self.cprobability


class BaseAlgorithmTestCase(unittest.TestCase):

    def setUp(self, **kwargs):
        self.base_algorithm = base.BaseAlgorithm(**kwargs)

    def tearDown(self):
        del self.base_algorithm

    def test_initialise(self):
        self.assertFalse(self.base_algorithm.initialise()[1])

    def test_iterate(self):
        self.assertTrue(self.base_algorithm.iterate({})[1])

    @mock.patch('MTfit.algorithms.base.BaseAlgorithm.get_sampling_model')
    def test___init__(self, get_sampling_model):
        self.base_algorithm = base.BaseAlgorithm()
        self.assertFalse(self.base_algorithm.dc)
        self.assertEqual(self.base_algorithm.number_samples, 10000)
        self.assertFalse(self.base_algorithm.mcmc)
        self.assertFalse(self.base_algorithm.basic_cdc)
        self.assertFalse(self.base_algorithm.quality_check)
        self.assertFalse(self.base_algorithm.generate)
        self.assertEqual(self.base_algorithm.number_events, 1)
        self.assertFalse(self.base_algorithm._model)
        get_sampling_model.assert_called_once_with({}, False, True)

    @mock.patch('MTfit.algorithms.base.BaseAlgorithm.get_sampling_model')
    def test___init___kwargs(self, get_sampling_model):
        self.base_algorithm = base.BaseAlgorithm(basic_cdc=3,
                                                 number_events=5,
                                                 sample_distribution=7,
                                                 file_sample=True,
                                                 file_safe=False)
        self.assertFalse(self.base_algorithm.dc)
        self.assertEqual(self.base_algorithm.number_samples, 10000)
        self.assertFalse(self.base_algorithm.mcmc)
        self.assertEqual(self.base_algorithm.basic_cdc, 3)
        self.assertFalse(self.base_algorithm.quality_check)
        self.assertFalse(self.base_algorithm.generate)
        self.assertEqual(self.base_algorithm.number_events, 5)
        self.assertEqual(self.base_algorithm._model, 7)
        get_sampling_model.assert_called_once_with({'basic_cdc': 3,
                                                    'number_events': 5,
                                                    'sample_distribution': 7},
                                                   True,
                                                   False)

    @unittest.expectedFailure
    @mock.patch('MTfit.algorithms.base.get_extensions')
    def test_get_sampling_model_model(self, get_extensions):
        raise NotImplementedError()

    def test_max_value(self):
        self.assertEqual(self.base_algorithm.max_value(), 'BaseAlgorithm has no max_value')

    def test_random_sample_generate(self):
        self.base_algorithm.generate = True
        self.assertFalse(self.base_algorithm.random_sample())

    @mock.patch('MTfit.algorithms.base.BaseAlgorithm.random_dc')
    def test_random_sample_random_dc(self, random_dc):
        self.base_algorithm.dc = True
        random_dc.return_value = 5
        self.assertEqual(self.base_algorithm.random_sample(), 5)
        random_dc.assert_called_once_with()

    def test_random_sample_random_basic_cdc(self):
        self.base_algorithm.dc = False
        self.base_algorithm.basic_cdc = True
        self.base_algorithm.random_basic_cdc = mock.MagicMock(return_value=5)
        self.assertEqual(self.base_algorithm.random_sample(), 5)
        self.base_algorithm.random_basic_cdc.assert_called_once_with()

    def test_random_sample_random_model(self):
        self.base_algorithm.dc = False
        self.base_algorithm.basic_cdc = False
        self.base_algorithm._model = True
        self.base_algorithm.random_model = mock.MagicMock(return_value=5)
        self.assertEqual(self.base_algorithm.random_sample(), 5)
        self.base_algorithm.random_model.assert_called_once_with(self.base_algorithm.number_samples)

    def test_random_sample_random_mt(self):
        self.base_algorithm.dc = False
        self.base_algorithm.basic_cdc = False
        self.base_algorithm._model = False
        self.base_algorithm.random_mt = mock.MagicMock(return_value=5)
        self.assertEqual(self.base_algorithm.random_sample(), 5)
        self.base_algorithm.random_mt.assert_called_once_with()

    @unittest.expectedFailure
    def test_output(self):
        raise NotImplementedError()

    def test_clvd_sampling(self):
        self.tearDown()
        self.setUp(sample_distribution='clvd')
        if sys.version_info.major > 2:
            self.assertEqual(self.base_algorithm.random_model.__name__, self.base_algorithm.random_clvd.__name__)
        else:
            self.assertEqual(self.base_algorithm.random_model, self.base_algorithm.random_clvd)

    def test_random_mt(self):
        self.assertTrue(self.base_algorithm.random_mt().shape,
                        (6, self.base_algorithm.number_samples))

    @mock.patch('MTfit.algorithms.base.logger')
    def test_random_dc_no_cython(self, logger):
        with PythonOnly():
            self.assertTrue(self.base_algorithm.random_dc().shape,
                            (6, self.base_algorithm.number_samples))
            logger.info.assert_called_once_with(C_EXTENSION_FALLBACK_LOG_MSG)

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.algorithms.base.logger')
    def test_random_dc_cython(self, logger):
        self.assertTrue(self.base_algorithm.random_dc().shape,
                        (6, self.base_algorithm.number_samples))
        logger.info.assert_not_called()

    def test_random_clvd(self):
        self.assertTrue(self.base_algorithm.random_clvd().shape,
                        (6, self.base_algorithm.number_samples))

    def test_random_type(self):
        self.base_algorithm.number_samples = 1
        self.assertAlmostEqual(
            self.base_algorithm.random_type(np.array([[1], [1], [1]]))[0, 0], 1/np.sqrt(3))

    def test_random_orthogonal_eigenvectors(self):
        self.base_algorithm.number_samples = 1
        [a, b, c] = self.base_algorithm.random_orthogonal_eigenvectors()
        self.assertAlmostEqual(c[0, 0], np.cross(a.transpose(), b.transpose()).transpose()[0, 0])
        self.assertAlmostEqual(c[1, 0], np.cross(a.transpose(), b.transpose()).transpose()[1, 0])
        self.assertAlmostEqual(c[2, 0], np.cross(a.transpose(), b.transpose()).transpose()[2, 0])

    def test_eigenvectors_mt_2_mt6(self):
        self.base_algorithm.number_samples = 1
        self.assertAlmostEqual(self.base_algorithm.eigenvectors_mt_2_mt6(np.array(
            [[1], [1], [1]]), *self.base_algorithm.random_orthogonal_eigenvectors())[0, 0], 1/np.sqrt(3))

    @unittest.skipIf(*C_EXTENSIONS)
    @mock.patch('MTfit.algorithms.base.logger')
    def test_6sphere_random_mt_cython(self, logger):
        self.base_algorithm.get_sampling_model = MethodType(base._6sphere_random_mt, self.base_algorithm)
        res = self.base_algorithm.get_sampling_model()
        logger.info.assert_not_called()
        self.assertEqual(res.shape, (6, self.base_algorithm.number_samples))
        self.assertAlmostEqual(np.prod(np.sum(np.multiply(res, res), axis=0)), 1)

    @mock.patch('MTfit.algorithms.base.logger')
    def test_6sphere_random_mt_no_cython(self, logger):
        self.base_algorithm.get_sampling_model = MethodType(base._6sphere_random_mt, self.base_algorithm)
        with PythonOnly():
            res = self.base_algorithm.get_sampling_model()
            logger.info.assert_called_once_with(C_EXTENSION_FALLBACK_LOG_MSG)
            self.assertEqual(res.shape, (6, self.base_algorithm.number_samples))
            self.assertAlmostEqual(np.prod(np.sum(np.multiply(res, res), axis=0)), 1)

    @mock.patch('MTfit.algorithms.base.logger')
    def test_6sphere_random_mt_int(self, logger):
        with PythonOnly():
            res = base._6sphere_random_mt(10)
            logger.info.assert_called_once_with(C_EXTENSION_FALLBACK_LOG_MSG)
            self.assertEqual(res.shape, (6, 10))
            self.assertAlmostEqual(np.prod(np.sum(np.multiply(res, res), axis=0)), 1)



import unittest
import os
import glob
import sys
import tempfile
import shutil

import numpy as np

from MTfit.sampling import Sample
from MTfit.sampling import FileSample
from MTfit.probability.probability import LnPDF


class SampleTestCase(unittest.TestCase):

    def setUp(self):
        self.Sample = Sample()

    def tearDown(self):
        del self.Sample

    def test_append(self):
        moment_tensors = np.matrix([[2], [1], [2], [1], [2], [1]])
        ln_pdf = LnPDF(np.matrix([[1], [2]]))
        self.Sample.append(moment_tensors, ln_pdf, 1)
        self.assertTrue((self.Sample.moment_tensors[:, 0] == np.matrix([[2], [1], [2], [1], [2], [1]])).all())
        self.assertEqual(self.Sample.n, 1)
        self.assertEqual(self.Sample.ln_pdf[0, 0], 1.)
        moment_tensors = np.matrix([[2, 3], [1, 5], [2, 1], [1, 1], [2, 2], [1, 1]])
        ln_pdf = LnPDF(np.matrix([[1, 0], [2, 1]]))
        self.Sample.append(moment_tensors, ln_pdf, 2)
        self.assertTrue((self.Sample.moment_tensors[:, 0:3] == np.matrix([[2, 2, 3], [1, 1, 5], [2, 2, 1], [1, 1, 1], [2, 2, 2], [1, 1, 1]])).all())
        moment_tensors = np.matrix([[2], [1], [2], [1], [2], [1]])
        ln_pdf = LnPDF(np.matrix([[0], [0]]))
        self.Sample.append(moment_tensors, ln_pdf, 1)
        self.assertTrue((self.Sample.moment_tensors[:, 0:3] == np.matrix([[2, 2, 3], [1, 1, 5], [2, 2, 1], [1, 1, 1], [2, 2, 2], [1, 1, 1]])).all())
        self.Sample.append(np.matrix([[], [], [], [], [], []]), np.matrix([[]]), 1)
        self.assertTrue((self.Sample.moment_tensors[:, 0:3] == np.matrix([[2, 2, 3], [1, 1, 5], [2, 2, 1], [1, 1, 1], [2, 2, 2], [1, 1, 1]])).all())
        self.assertEqual(self.Sample.n, 5)
        # test multiple events
        self.tearDown()
        self.Sample = Sample(initial_sample_size=1, number_events=2)
        moment_tensors = [np.matrix([[2], [1], [2], [1], [2], [1]]), np.matrix([[2], [1], [2], [1], [2], [1]])]
        ln_pdf = LnPDF(np.matrix([[1]]))
        self.Sample.append(moment_tensors, ln_pdf, 1)
        self.assertTrue((self.Sample.moment_tensors[:, 0] == np.matrix([[2], [1], [2], [1], [2], [1], [2], [1], [2], [1], [2], [1]])).all())
        # test scale_factor - multiple events
        self.tearDown()
        self.Sample = Sample(initial_sample_size=1, number_events=2)
        moment_tensors = [np.matrix([[2], [1], [2], [1], [2], [1]]), np.matrix([[2], [1], [2], [1], [2], [1]])]
        ln_pdf = LnPDF(np.matrix([[1]]))
        scale_factor = np.array([{'mu': np.array([[[2, 3], [3, 2]]]), 'sigma': np.array([[1, 1], [1, 1]])}])
        self.assertFalse(self.Sample.__dict__.__contains__('scale_factor'))
        self.Sample.append(moment_tensors, ln_pdf, 1, scale_factor)
        self.assertTrue((self.Sample.moment_tensors[:, 0] == np.matrix([[2], [1], [2], [1], [2], [1], [2], [1], [2], [1], [2], [1]])).all())
        self.assertTrue(self.Sample.__dict__.__contains__('scale_factor'))
        self.assertTrue((self.Sample.scale_factor[0]['mu'] == np.array([[2, 3], [3, 2]])).all())
        self.tearDown()
        self.Sample = Sample(initial_sample_size=3)
        # test initial samples
        moment_tensors = np.matrix([[2], [1], [2], [1], [2], [1]])
        ln_pdf = LnPDF(np.matrix([[1], [2]]))
        self.Sample.append(moment_tensors, ln_pdf, 1)
        self.assertTrue((self.Sample.moment_tensors[:, 0] == np.matrix([[2], [1], [2], [1], [2], [1]])).all())
        self.assertEqual(self.Sample.n, 1)
        self.assertEqual(self.Sample.ln_pdf[0, 0], 1.)
        moment_tensors = np.matrix([[2, 3], [1, 5], [2, 1], [1, 1], [2, 2], [1, 1]])
        ln_pdf = LnPDF(np.matrix([[1, 0], [2, 1]]))
        self.Sample.append(moment_tensors, ln_pdf, 2)
        self.assertTrue((self.Sample.moment_tensors[:, 0:3] == np.matrix([[2, 2, 3], [1, 1, 5], [2, 2, 1], [1, 1, 1], [2, 2, 2], [1, 1, 1]])).all())

        moment_tensors = np.matrix([[2], [1], [2], [1], [2], [1]])
        ln_pdf = LnPDF(np.matrix([[1], [2]]))
        self.Sample.append(moment_tensors, ln_pdf, 1)
        self.assertTrue((self.Sample.moment_tensors[:, 0:4] == np.matrix([[2, 2, 3, 2], [1, 1, 5, 1], [2, 2, 1, 2], [1, 1, 1, 1], [2, 2, 2, 2], [1, 1, 1, 1]])).all())

    def test_output(self):
        self.test_append()
        self.assertTrue((self.Sample.output()[0]['moment_tensor_space'] == np.matrix([[2, 2, 3, 2], [1, 1, 5, 1], [2, 2, 1, 2], [1, 1, 1, 1], [2, 2, 2, 2], [1, 1, 1, 1]])).all())
        output_ln_pdf = np.sum(np.exp(np.matrix([[1., 1., 0., 1.], [2., 2., 1., 2.]])), 0)
        output_ln_pdf /= np.sum(output_ln_pdf)
        self.assertAlmostEqual(self.Sample.output()[0]['probability'][0], output_ln_pdf[0, 0])
        self.assertAlmostEqual(self.Sample.output()[0]['probability'][1], output_ln_pdf[0, 1])
        self.assertAlmostEqual(self.Sample.output()[0]['probability'][2], output_ln_pdf[0, 2])
        self.assertAlmostEqual(self.Sample.output()[0]['probability'][3], output_ln_pdf[0, 3])
        self.assertEqual(self.Sample.output()[0]['dV'], 1)
        output_ln_pdf = np.sum(np.exp(np.matrix([[1., 1., 0., 1.], [2., 2., 1., 2.]])), 0)
        self.assertAlmostEqual(self.Sample.output(False)[0]['probability'][0], output_ln_pdf[0, 0])
        self.assertAlmostEqual(self.Sample.output(False)[0]['probability'][1], output_ln_pdf[0, 1])
        self.assertAlmostEqual(self.Sample.output(False)[0]['probability'][2], output_ln_pdf[0, 2])
        self.assertAlmostEqual(self.Sample.output(False)[0]['probability'][3], output_ln_pdf[0, 3])
        self.assertEqual(self.Sample.output(False)[0]['dV'], 1)
        self.tearDown()
        self.Sample = Sample(initial_sample_size=1, number_events=2)
        moment_tensors = [np.matrix([[2], [1], [2], [1], [2], [1]]), np.matrix([[3], [1], [2], [1], [2], [1]])]
        ln_pdf = LnPDF(np.matrix([[1]]))
        self.Sample.append(moment_tensors, ln_pdf, 1)
        self.assertTrue((self.Sample.moment_tensors[:, 0] == np.matrix([[2], [1], [2], [1], [2], [1], [3], [1], [2], [1], [2], [1]])).all())
        self.assertTrue((self.Sample.output()[0]['moment_tensor_space_1'] == np.matrix([[2], [1], [2], [1], [2], [1]])).all())
        self.assertTrue((self.Sample.output()[0]['moment_tensor_space_2'] == np.matrix([[3], [1], [2], [1], [2], [1]])).all())
        self.assertTrue((self.Sample.output()[0]['probability'] == (np.array([1]))).all(), str(self.Sample.output()[0]['probability']))
        self.assertEqual(self.Sample.output()[0]['dV'], 1)
        self.tearDown()
        self.Sample = Sample(initial_sample_size=1, number_events=2)
        moment_tensors = [np.matrix([[2], [1], [2], [1], [2], [1]]), np.matrix([[2], [1], [2], [1], [2], [1]])]
        ln_pdf = LnPDF(np.matrix([[1.]]))
        scale_factor = np.array([{'mu': np.array([[[2, 3], [3, 2]]]), 'sigma': np.array([[1, 1], [1, 1]])}])
        self.assertFalse(self.Sample.__dict__.__contains__('scale_factor'))
        self.Sample.append(moment_tensors, ln_pdf, 1, scale_factor)
        self.assertTrue((self.Sample.moment_tensors[:, 0] == np.matrix([[2], [1], [2], [1], [2], [1], [2], [1], [2], [1], [2], [1]])).all())
        self.assertTrue(self.Sample.__dict__.__contains__('scale_factor'))
        self.assertTrue((self.Sample.scale_factor[0]['mu'] == np.array([[2, 3], [3, 2]])).all())
        self.assertTrue((self.Sample.output()[0]['scale_factors'][0]['mu'] == np.array([[[2, 3], [3, 2]]])).all(), str(self.Sample.output()[0]['scale_factors'][0]['mu']))
        self.tearDown()
        self.setUp()
        self.test_append()
        out, out_str = self.Sample.output(convert=True)
        self.assertTrue('dV' in out.keys())
        self.assertTrue('moment_tensor_space' in out.keys())
        self.assertTrue('probability' in out.keys())
        self.assertTrue('ln_pdf' in out.keys())
        self.assertTrue('g' in out.keys())


class FileSampleTestCase(unittest.TestCase):
    def setUp(self):
        self.cwd = os.getcwd()
        if sys.version_info >= (3, 0):
            self.tempdir = tempfile.TemporaryDirectory()
            os.chdir(self.tempdir.name)
        else:
            self.tempdir = tempfile.mkdtemp()
            os.chdir(self.tempdir)
        self.FileSample = FileSample('test')

    def tearDown(self):
        output_files = glob.glob('*in_progress*.mat*')
        for filename in output_files:
            try:
                os.remove(filename)
            except Exception:
                pass
        output_files = glob.glob('*.mat*')
        for filename in output_files:
            try:
                os.remove(filename)
            except Exception:
                pass
        os.chdir(self.cwd)
        if sys.version_info >= (3, 0):
            self.tempdir.cleanup()
        else:
            try:
                shutil.rmtree(self.tempdir)
            except:
                pass

    def test___init__(self):
        self.assertEqual(self.FileSample.fname, 'test_in_progress.mat')
        self.assertEqual(self.FileSample.n, 0)
        self.assertEqual(self.FileSample._i, 1)

    def test_recover(self):
        self.test_append()
        self.__setattr__('FileSample', FileSample('test'))
        self.assertEqual(self.FileSample._i, 3)
        self.assertEqual(self.FileSample.n, 50)
        self.assertEqual(self.FileSample.non_zero_samples, 2)

    def test_append(self):
        self.assertEqual(self.FileSample.fname, 'test_in_progress.mat')
        self.assertEqual(self.FileSample.n, 0)
        self.FileSample.append(np.matrix([[1], [2], [1], [2], [1], [1]]), np.matrix([1]), 10)
        self.assertEqual(self.FileSample.n, 10)
        self.assertTrue(os.path.exists(self.FileSample.fname))
        from hdf5storage import loadmat
        x = loadmat(self.FileSample.fname)
        self.assertEqual(sorted(x.keys()), ['LnPDF_1', 'MTSpace_1', 'i', 'n', 'non_zero_samples'])
        self.assertEqual(x['n'], 10)
        self.FileSample.append(np.matrix([[1], [2], [1], [2], [1], [1]]), np.matrix([1]), 20)
        self.assertEqual(self.FileSample.n, 30)
        self.assertTrue(os.path.exists(self.FileSample.fname))
        x = loadmat(self.FileSample.fname)
        self.assertEqual(sorted(x.keys()), ['LnPDF_1', 'LnPDF_2', 'MTSpace_1', 'MTSpace_2', 'i', 'n', 'non_zero_samples'])
        self.assertEqual(x['n'], 30)
        self.FileSample.append(np.matrix([[1], [2], [1], [2], [1], [1]]), np.matrix([-np.inf]), 20)
        self.assertEqual(self.FileSample.n, 50)
        x = loadmat(self.FileSample.fname)
        self.assertEqual(sorted(x.keys()), ['LnPDF_1', 'LnPDF_2', 'MTSpace_1', 'MTSpace_2', 'i', 'n', 'non_zero_samples'])
        self.assertEqual(x['n'], 50)
        self.tearDown()
        self.setUp()
        self.assertEqual(self.FileSample.fname, 'test_in_progress.mat')
        self.assertEqual(self.FileSample.n, 0)
        self.FileSample.append(np.matrix([[1], [2], [1], [2], [1], [1]]), np.matrix([1]), 10, [{'mu': np.array([[1, 2], [2, 1]]), 'sigma': np.array([[0, 0.1], [0.1, 0]])}])
        self.assertEqual(self.FileSample.n, 10)
        self.assertTrue(os.path.exists(self.FileSample.fname))
        from hdf5storage import loadmat
        x = loadmat(self.FileSample.fname)
        self.assertEqual(sorted(x.keys()), ['LnPDF_1', 'MTSpace_1', 'i', 'n', 'non_zero_samples', 'scale_factor_1'])
        self.assertEqual(x['n'], 10)
        self.FileSample.append(np.matrix([[1], [2], [1], [2], [1], [1]]), np.matrix([1]), 20, [{'mu': np.array([[1, 2], [2, 1]]), 'sigma': np.array([[0, 0.1], [0.1, 0]])}])
        self.assertEqual(self.FileSample.n, 30)
        self.assertTrue(os.path.exists(self.FileSample.fname))
        x = loadmat(self.FileSample.fname)
        self.assertEqual(sorted(x.keys()), ['LnPDF_1', 'LnPDF_2', 'MTSpace_1', 'MTSpace_2', 'i', 'n', 'non_zero_samples', 'scale_factor_1', 'scale_factor_2'])
        self.assertEqual(x['n'], 30)
        self.FileSample.append(np.matrix([[1], [2], [1], [2], [1], [1]]), np.matrix([-np.inf]), 20, [{'mu': np.array([[1, 2], [2, 1]]), 'sigma': np.array([[0, 0.1], [0.1, 0]])}])
        self.assertEqual(self.FileSample.n, 50)
        x = loadmat(self.FileSample.fname)
        self.assertEqual(sorted(x.keys()), ['LnPDF_1', 'LnPDF_2', 'MTSpace_1', 'MTSpace_2', 'i', 'n', 'non_zero_samples', 'scale_factor_1', 'scale_factor_2'])
        self.assertEqual(x['n'], 50)

    def test_output(self):
        self.test_append()
        out, out_str = self.FileSample.output(convert=True)
        self.assertTrue('dV' in out.keys())
        self.assertTrue('moment_tensor_space' in out.keys())
        self.assertTrue('probability' in out.keys())
        self.assertTrue('ln_pdf' in out.keys())
        self.assertTrue('g' in out.keys())
        self.assertEqual(out['moment_tensor_space'].shape, (6, 2))
        self.assertEqual(out['probability'].shape, (2, ))



"""
probability.py
**************

Probability module for MTfit, handles all probability calculations and
contains the base LnPDF class for acting on probabilities.

LnPDF object acts as a wrapper around a numpy matrix, allowing some
additional pdf specific operations.
"""


# **Restricted:  For Non-Commercial Use Only**
# This code is protected intellectual property and is available solely for teaching
# and non-commercially funded academic research purposes.
#
# Applications for commercial use should be made to Schlumberger or the University of Cambridge.


import warnings
import gc
import sys
import operator
import logging

import numpy as np
from scipy.stats import norm as gaussian
from scipy.stats import beta
from scipy.special import erf

from ..utilities import C_EXTENSION_FALLBACK_LOG_MSG

logger = logging.getLogger('MTfit.probability')


try:
    from . import cprobability
except ImportError:
    cprobability = None
except Exception:
    logger.exception('Error importing c extension')
    cprobability = None

# Set to prevent numpy warnings
np.seterr(divide='print', invalid='print')

# Set flags for running with/without the c library

# Flag for testing Cython based C library functions even if they would be skipped
# e.g. polarity_ln_pdf on windows due to missing functions in math.h
_C_LIB_TESTS = False

# Value of a very small non-zero number for handling zero error values
_SMALL_NUMBER = 0.000000000000000000000001

# TODO - tidy and refactor code to avoid duplication


def _6sphere_prior(g, d):
    """
    6-sphere prior

    Returns 1 as samples are generated directly from the prior

    Other priors can be used either by non-uniform sampling in the prior
    and using the prior as the correction term in the bayesian evidence
    calculation, or by leaving the prior as 1 and sampling from the prior
    """
    return 1.


def polarity_ln_pdf(a, mt, sigma, incorrect_polarity_probability=0.0, _use_c=None):
    """
    Calculate the probability of a positive polarity

    Calculates the probability of a positive polarity observation given
    a theoretical amplitude X and a fractional uncertainty sigma.

    The derivation of this pdf for the polarity observation can be seen in:
        Pugh, D. J., White, R. S., & Christie, P. A. F., 2016.
        A Bayesian method for microseismic source inversion,
        Geophysical Journal International , 206(2), 1009-1038.

    Args:
        a: np.array - 3 dimensional numpy array of station coefficients
        mt: np.array - 2 dimensional numpy array of moment tensor 6
                vector samples
        sigma: np.array - 1 dimensional array of fractional uncertainties.

    Optional Args:
        incorrect_polarity_probability: float (default=0) - probability
                of a receiver orientation error, flipping the polarity dist.

    Returns:
        np.array - probabilities for each moment tensor sample
    """
    # Check inputs and expcected dimensions
    if not isinstance(sigma, np.ndarray) or sigma.ndim != 1:
        raise TypeError('Variable: sigma is expected to be a one-dimensional numpy array')
    if not isinstance(a, np.ndarray) or a.ndim != 3:
        raise TypeError('Variable: a is expected to be a three-dimensional numpy array')
    if not isinstance(mt, np.ndarray) or mt.ndim != 2:
        raise TypeError('Variable: mt is expected to be a two-dimensional numpy array of moment tensor six vectors')
    # For fractional uncertainties that are zero make them really small
    # to avoid zero divison errors
    sigma[sigma == 0] = _SMALL_NUMBER
    completed = False
    if cprobability and (_use_c is None or _use_c):
        try:
            if sys.platform.startswith('win') and not _use_c:
                # Raise an exception due to windows C_LIB using VS2008 VC math.h
                # which has no erf
                raise Exception('Windows')
            # Run using C library
            # Handle incorrect polarity probability
            if isinstance(incorrect_polarity_probability, (float, int)):
                incorrect_polarity_probability = np.ones(sigma.shape)*incorrect_polarity_probability
            incorrect_polarity_probability = np.squeeze(incorrect_polarity_probability)
            if incorrect_polarity_probability.ndim != 1:
                raise TypeError('Variable: incorrect_polarity_probability is expected to be a one-dimensional numpy array or a float')
            # Check types and convert to the correct types in place
            if a.dtype != np.float64:
                a = a.astype(np.float64, copy=False)
            if mt.dtype != np.float64:
                mt = mt.astype(np.float64, copy=False)
            if sigma.dtype != np.float64:
                sigma = sigma.astype(np.float64, copy=False)
            if incorrect_polarity_probability.dtype != np.float64:
                incorrect_polarity_probability = incorrect_polarity_probability.astype(np.float64,
                                                                                       copy=False)
            # Make sure moment tensors are C contiguous
            if not mt.flags['C_CONTIGUOUS']:
                mt = np.ascontiguousarray(mt)
            # Calculate log of pdf
            ln_p = cprobability.polarity_ln_pdf(a, mt, sigma, incorrect_polarity_probability)
            completed = True
        except Exception as e:
            # Run using python
            # Testing C code
            logger.exception('WARNING: Error running cython code, resorting to python code')
            if _C_LIB_TESTS:
                raise e
    else:
        logger.info(C_EXTENSION_FALLBACK_LOG_MSG)
    if not completed:
        # Check moment tensor shape is correct
        if mt.shape[0] != a.shape[-1]:
            mt = mt.transpose()
        # Calcuate Theoretical Amplitude for moment tensors
        X = np.tensordot(a, mt, 1)
        # Expand errors and incorrect_polarity_probability dimensions to be
        # expected
        while sigma.ndim < 3:
            sigma = np.expand_dims(sigma, 1)
        if not isinstance(incorrect_polarity_probability, (float, int)):
            incorrect_polarity_probability = np.array(incorrect_polarity_probability)
        if isinstance(incorrect_polarity_probability, np.ndarray):
            while incorrect_polarity_probability.ndim < 3:
                incorrect_polarity_probability = np.expand_dims(
                    incorrect_polarity_probability, 1)
        # Calculate probability
        ln_p = np.log((0.5 * (1 + erf(X / (np.sqrt(2) * sigma)))) *
                      (1 - incorrect_polarity_probability) +
                      (0.5 * (1 + erf(-X / (np.sqrt(2) * sigma)))) *
                      incorrect_polarity_probability)
        # Set NaNs to zero probability
        if isinstance(ln_p, np.ndarray):
            ln_p[np.isnan(ln_p)] = -np.inf
        # Combine probabilities
        try:
            ln_p = cprobability.ln_prod(ln_p)
        except Exception:
            if cprobability:
                logger.exception('WARNING: Error running cython code, resorting to python code')
            ln_p = np.sum(ln_p, 0)
    if isinstance(ln_p, np.ndarray):
        ln_p[np.isnan(ln_p)] = -np.inf
    return ln_p


def polarity_probability_ln_pdf(a, mt, positive_probability, negative_probability, incorrect_polarity_probability=0.0, _use_c=None):
    """
    Calculate the probability of a given theoretical amplitude giving an
    observed polarity probability.

    Calculates the probability of a polarity probability observation
    given a theoretical amplitude X.

    The derivation of this pdf for the polarity probability observation
    can be seen in:
        Pugh, D. J., White, R. S., & Christie, P. A. F., 2016.
        Automatic Bayesian polarity determination,
        Geophysical Journal International , 206(1), 275-291.

    Args
        a: np.array - 3 dimensional numpy array of station coefficients
        mt: np.array - 2 dimensional numpy array of moment tensor 6 vector
                samples
        positive_probability: float/np.array - the probability of a positive polarity
                observation. Needs to be same length as mt, or a scalar.
        negative_probability: float/np.array - the probability of a negative polarity
                observation. Needs to be same length as mt, or a scalar.

    Optional Args:
        incorrect_polarity_probability: float (default=0) - probability of a
                receiver orientation error, flipping the polarity dist.

    Returns
        float/np.array -  probabilities for each moment tensor sample
    """
    # Check inputs and expcected dimensions
    if isinstance(positive_probability, (float, int)):
        positive_probability = np.array(positive_probability)
    if isinstance(negative_probability, (float, int)):
        negative_probability = np.array(negative_probability)
    if not isinstance(positive_probability, np.ndarray) or positive_probability.ndim != 1:
        raise TypeError(
            'Variable: positive_probability is expected to be a one-dimensional numpy array')
    if not isinstance(negative_probability, np.ndarray) or negative_probability.ndim != 1:
        raise TypeError(
            'Variable: negative_probability is expected to be a one-dimensional numpy array')
    if not isinstance(a, np.ndarray) or a.ndim != 3:
        raise TypeError('Variable: a is expected to be a three-dimensional numpy array')
    if not isinstance(mt, np.ndarray) or mt.ndim != 2:
        raise TypeError('Variable: mt is expected to be a two-dimensional numpy array of moment tensor six vectors')
    completed = False
    if cprobability and (_use_c is None or _use_c):
        try:
            # Run using C library
            # Handle incorrect polarity probability
            if isinstance(incorrect_polarity_probability, (float, int)):
                incorrect_polarity_probability = np.ones(positive_probability.shape)*incorrect_polarity_probability
            incorrect_polarity_probability = np.squeeze(incorrect_polarity_probability)
            if incorrect_polarity_probability.ndim != 1:
                raise TypeError('Variable: incorrect_polarity_probability is expected to be a one-dimensional numpy array or a float')
            # Check types and convert to the correct types in place
            if a.dtype != np.float64:
                a = a.astype(np.float64, copy=False)
            if mt.dtype != np.float64:
                mt = mt.astype(np.float64, copy=False)
            if incorrect_polarity_probability.dtype != np.float64:
                incorrect_polarity_probability = incorrect_polarity_probability.astype(np.float64,
                                                                                       copy=False)
            if positive_probability.dtype != np.float64:
                positive_probability = positive_probability.astype(np.float64, copy=False)
            if negative_probability.dtype != np.float64:
                negative_probability = negative_probability.astype(np.float64, copy=False)
            # Make sure moment tensors are C contiguous
            if not mt.flags['C_CONTIGUOUS']:
                mt = np.ascontiguousarray(mt)
            # Calculate log of pdf
            ln_p = cprobability.polarity_probability_ln_pdf(a, mt, positive_probability, negative_probability,
                                                            incorrect_polarity_probability)
            completed = True
        except Exception as e:
            # Run using python
            # Testing C code
            logger.exception('WARNING: Error running cython code, resorting to python code')
            if _C_LIB_TESTS:
                raise e
    else:
        logger.info(C_EXTENSION_FALLBACK_LOG_MSG)
    if not completed:
        if mt.shape[0] != a.shape[-1]:
            mt = mt.transpose()
        # Calculate theoretical amplitude
        X = np.tensordot(a, mt, 1)
        # Expand errors and incorrect_polarity_probability dimensions to be
        # expected
        while positive_probability.ndim < 3:
            positive_probability = np.expand_dims(positive_probability, 1)
        while negative_probability.ndim < 3:
            negative_probability = np.expand_dims(negative_probability, 1)
        if not isinstance(incorrect_polarity_probability, (float, int)):
            incorrect_polarity_probability = np.array(
                incorrect_polarity_probability)
        if isinstance(incorrect_polarity_probability, np.ndarray):
            while incorrect_polarity_probability.ndim < 3:
                incorrect_polarity_probability = np.expand_dims(
                    incorrect_polarity_probability, 1)
        ln_p = np.log(
            ((heaviside(X) * positive_probability) + (heaviside(-X) * negative_probability)) *
            (1-incorrect_polarity_probability) +
            ((heaviside(X) * negative_probability) + (heaviside(-X) * positive_probability)) *
            incorrect_polarity_probability)
        # Set NaNs to zero probability
        if isinstance(ln_p, np.ndarray):
            ln_p[np.isnan(ln_p)] = -np.inf
        # Combine probabilities
        try:
            ln_p = cprobability.ln_prod(ln_p)
        except Exception:
            ln_p = np.sum(ln_p, 0)
    if isinstance(ln_p, np.ndarray):
        ln_p[np.isnan(ln_p)] = -np.inf
    return ln_p


def amplitude_ratio_ln_pdf(ratio, mt, a_x, a_y, percentage_error_x, percentage_error_y, _use_c=None):
    """
    Calculate the probability of a given theoretical amplitude ratio giving
    an observed ratio

    Calculates the probability of an amplitude ratio observation given a
    theoretical amplitude ratio and uncertainties on the numerator and
    denominator

    The derivation of this pdf for the amplitude ratio observation can be
    seen in:
        Pugh, D. J., White, R. S., & Christie, P. A. F., 2016.
        A Bayesian method for microseismic source inversion,
        Geophysical Journal International , 206(2), 1009-1038.

    Args
        ratio: np.array - 1 dimensional numpy array of observed ratio
                values
        mt: np.array - 2 dimensional numpy array of moment tensor 6
                vector samples
        a_x: np.array - 3 dimensional numpy array of station coefficients
                for the numerator
        a_y: np.array - 3 dimensional numpy array of station coefficients
                for the denominator
        percentage_error_x: np.array - 1 dimensional numpy array of
                percentage errors for the numerator
        percentage_error_y: np.array - 1 dimensional numpy array of
                percentage errors for the denominator

    Returns
        float/np.array -  probabilities for each moment tensor sample
    """
    # Check inputs and expcected dimensions
    if not isinstance(a_x, np.ndarray) or a_x.ndim != 3:
        raise TypeError('Variable a_x is expected to be a three-dimensional numpy array')
    if not isinstance(a_y, np.ndarray) or a_y.ndim != 3:
        raise TypeError('Variable: a_y is expected to be a three-dimensional numpy array')
    if not isinstance(mt, np.ndarray) or mt.ndim != 2:
        raise TypeError(
            'Variable: mt is expected to be a two-dimensional numpy array of moment tensor six vectors')
    if not isinstance(ratio, np.ndarray) or ratio.ndim != 1:
        raise TypeError(
            'Variable: ratio is expected to be a one-dimensional numpy array')
    if not isinstance(percentage_error_x, np.ndarray) or percentage_error_x.ndim != 1:
        raise TypeError(
            'Variable: percentage_error_x is expected to be a one-dimensional numpy array')
    if not isinstance(percentage_error_y, np.ndarray) or percentage_error_y.ndim != 1:
        raise TypeError(
            'Variable: percentage_error_y is expected to be a one-dimensional numpy array')
    # Handle zero errors by making them small to avoid zero division errors
    percentage_error_x[percentage_error_x == 0] = _SMALL_NUMBER
    percentage_error_y[percentage_error_y == 0] = _SMALL_NUMBER
    # Make sure the errors are positive
    percentage_error_x = np.abs(percentage_error_x)
    percentage_error_y = np.abs(percentage_error_y)
    completed = False
    if cprobability and (_use_c is None or _use_c):
        try:
            ln_p = cprobability.amplitude_ratio_ln_pdf(ratio, mt, a_x, a_y, percentage_error_x,
                                                       percentage_error_y)
            completed = True
        except Exception as e:
            # Run using python
            # Testing C code
            logger.exception('WARNING: Error running cython code, resorting to python code')
            if _C_LIB_TESTS:
                raise e
    else:
        logger.info(C_EXTENSION_FALLBACK_LOG_MSG)
    if not completed:
        # Calculate probability using Python code
        # Calculate means for numerator and denominator
        mu_x = np.tensordot(a_x, mt, 1)
        mu_y = np.tensordot(a_y, mt, 1)
        # Make sure they are positive
        mu_x = np.abs(mu_x)
        mu_y = np.abs(mu_y)
        # Expand errors and ratio values to be 3 dimensional
        while percentage_error_x.ndim < 3:
            percentage_error_x = np.expand_dims(percentage_error_x, 1)
        while percentage_error_y.ndim < 3:
            percentage_error_y = np.expand_dims(percentage_error_y, 1)
        while ratio.ndim < 3:
            ratio = np.expand_dims(ratio, 1)
        # Get error on numerator and denominator
        numerator_error = np.multiply(percentage_error_x, mu_x)
        denominator_error = np.multiply(percentage_error_y, mu_y)
        # Calculate log of pdf
        ln_p = np.log(ratio_pdf(ratio, mu_x, mu_y, numerator_error, denominator_error) +
                      ratio_pdf(-ratio, mu_x, mu_y, numerator_error, denominator_error))
        # Set NaNs to zero probability
        if isinstance(ln_p, np.ndarray):
            ln_p[np.isnan(ln_p)] = -np.inf
        # Combine probabilities
        try:
            ln_p = cprobability.ln_prod(ln_p)
        except Exception:
            if cprobability:
                logger.exception('WARNING: Error running cython code, resorting to python code')
            ln_p = np.sum(ln_p, 0)
    if isinstance(ln_p, np.ndarray):
        ln_p[np.isnan(ln_p)] = -np.inf
    return ln_p


def relative_amplitude_ratio_ln_pdf(x_1, x_2, mt_1, mt_2, a_1, a_2, percentage_error_1, percentage_error_2, _use_c=None):
    """
    Calculate the probability of a given theoretical relative amplitude
    ratio giving an observed ratio.

    Calculates the probability of an amplitude ratio observation given a
    theoretical amplitude ratio and uncertainties on the numerator and
    denominator.

    The derivation of this pdf for the relative amplitude observation
    can be seen in:
        Pugh, D. J., 2015.
        Bayesian Source Inversion of Microseismic Events,
        Ph.D. thesis, University of Cambridge.

    Args
        x: np.array - 1 dimensional numpy array of observed numerator
                values
        y: np.array - 1 dimensional numpy array of observed denominator
                values
        mt_1: np.array - 2 dimensional numpy array of moment tensor 6
                vector samples for numerator
        mt_2: np.array - 2 dimensional numpy array of moment tensor 6
                vector samples for denominator
        a_1: np.array - 3 dimensional numpy array of station coefficients
                for the numerator
        a_2: np.array - 3 dimensional numpy array of station coefficients
                for the denominator
        percentage_error_1: np.array - 1 dimensional numpy array of
            percentage errors for the numerator
        percentage_error_2: np.array - 1 dimensional numpy array of
            percentage errors for the denominator

    Returns
        (np.array, np.array. np.array) -  tuple of probabilities for
                each joint moment tensor sample, scale factors for the
                event and the uncertainties in the scale factor.
    """
    # Check inputs and expcected dimensions
    if not isinstance(a_1, np.ndarray) or a_1.ndim != 3:
        raise TypeError('Variable: a_1 is expected to be a three-dimensional numpy array')
    if not isinstance(a_2, np.ndarray) or a_2.ndim != 3:
        raise TypeError('a_2 is expected to be a three-dimensional numpy array')
    if not isinstance(mt_1, np.ndarray) or mt_1.ndim != 2:
        raise TypeError('mt_1 is expected to be a two-dimensional numpy array')
    if not isinstance(mt_2, np.ndarray) or mt_2.ndim != 2:
        raise TypeError('mt_2 is expected to be a two-dimensional numpy array')
    if not isinstance(x_1, np.ndarray) or x_1.ndim != 1:
        raise TypeError('x is expected to be a one-dimensional numpy array')
    if not isinstance(x_2, np.ndarray) or x_2.ndim != 1:
        raise TypeError('y is expected to be a one-dimensional numpy array')
    if not isinstance(percentage_error_1, np.ndarray) or percentage_error_1.ndim != 1:
        raise TypeError('percentage_error_1 is expected to be a one-dimensional numpy array')
    if not isinstance(percentage_error_2, np.ndarray) or percentage_error_2.ndim != 1:
        raise TypeError('percentage_error_2 is expected to be a one-dimensional numpy array')
    # For fractional uncertainties that are zero make them really small
    # to avoid zero divison errors
    percentage_error_1[percentage_error_1 == 0] = _SMALL_NUMBER
    percentage_error_2[percentage_error_2 == 0] = _SMALL_NUMBER
    # Make sure the errors are positive
    percentage_error_1 = np.abs(percentage_error_1)
    percentage_error_2 = np.abs(percentage_error_2)
    completed = False
    if cprobability and (_use_c is None or _use_c):
        try:
            # raise ValueError('C code returning incorrect result for probabilities')
            ln_p, scale, scale_uncertainty = cprobability.relative_amplitude_ratio_ln_pdf(np.ascontiguousarray(x_1),
                                                                                          np.ascontiguousarray(x_2),
                                                                                          np.ascontiguousarray(mt_1),
                                                                                          np.ascontiguousarray(mt_2),
                                                                                          np.ascontiguousarray(a_1),
                                                                                          np.ascontiguousarray(a_2),
                                                                                          np.ascontiguousarray(percentage_error_1),
                                                                                          np.ascontiguousarray(percentage_error_2))
            completed = True
        except Exception as e:
            # Run using python
            # Testing C code
            logger.exception('WARNING: Error running cython code, resorting to python code')
            if _C_LIB_TESTS:
                raise e
    else:
        logger.info(C_EXTENSION_FALLBACK_LOG_MSG)
    if not completed:
        # Calculate probability using Python code
        # Calculate the ratio
        ratio = np.divide(x_1, x_2)
        # Calculate means for numerator and denominator
        mu_x = np.tensordot(a_1, mt_1, 1)
        mu_y = np.tensordot(a_2, mt_2, 1)
        # Make sure they are positive
        mu_x = np.abs(mu_x)
        mu_y = np.abs(mu_y)
        # Expand errors and ratio values to be 3 dimensional
        while percentage_error_1.ndim < 3:
            percentage_error_1 = np.expand_dims(percentage_error_1, 1)
        while percentage_error_2.ndim < 3:
            percentage_error_2 = np.expand_dims(percentage_error_2, 1)
        while ratio.ndim < 3:
            ratio = np.expand_dims(ratio, 1)
        # Estimate scale factor
        scale, scale_uncertainty = scale_estimator(ratio, mu_x, mu_y,
                                                   percentage_error_1, percentage_error_2)
        # Multiply in the scale factor
        mu_x = np.multiply(scale, mu_x)
        # Calculate the numerator and denominator errors
        numerator_error = np.multiply(percentage_error_1, mu_x)
        denominator_error = np.multiply(percentage_error_2, mu_y)
        # Calculate the log pdf
        ln_p = np.log(ratio_pdf(ratio, mu_x, mu_y, numerator_error, denominator_error) +
                      ratio_pdf(-ratio, mu_x, mu_y, numerator_error, denominator_error))
        # Set NaNs to zero probability
        if isinstance(ln_p, np.ndarray):
            ln_p[np.isnan(ln_p)] = -np.inf
        # Combine probabilities
    try:
        ln_p = cprobability.ln_prod(ln_p)
    except Exception:
        ln_p = np.sum(ln_p, 0)
    if isinstance(ln_p, np.ndarray):
        ln_p[np.isnan(ln_p)] = -np.inf
    return ln_p, scale, scale_uncertainty


def scale_estimator(observed_ratio, mu_x, mu_y, percentage_error_x, percentage_error_y):
    """
    Estimate the scale factor between the two events given the observed_ratio
    and theoretical ratios.

    The derivation of this calculation can be found in:
        Pugh, D. J., 2015.
        Bayesian Source Inversion of Microseismic Events,
        Ph.D. thesis, University of Cambridge.

    Args
        observed_ratio: np.array - numpy array of observed_ratio ratio values
        mu_x: np.array - numpy array of mean values for the numerator
        mu_y: np.array - numpy array of station coefficients for the
            denominator
        percentage_error_x: np.array - numpy array of percentage errors
            for the numerator
        percentage_error_y: np.array - numpy array of percentage errors
            for the denominator

    Returns
        (np.array, np.array) -  tuple of means and standard deviations
                of the scale factor for each sample
    """
    # Expand arrays to be correct dimensions
    if isinstance(mu_x, np.ndarray) and mu_x.ndim == 3:
        while observed_ratio.ndim < 3:
            observed_ratio = np.expand_dims(observed_ratio, 1)
        while percentage_error_x.ndim < 3:
            percentage_error_x = np.expand_dims(percentage_error_x, 1)
        while percentage_error_y.ndim < 3:
            percentage_error_y = np.expand_dims(percentage_error_y, 1)
    # Calculate standard deviations for the numerator and denominator
    sx = np.multiply(np.array(percentage_error_x), mu_x)
    sy = np.multiply(np.array(percentage_error_y), mu_y)
    # Convert the observed_ratio values to an array
    observed_ratio = np.array(observed_ratio)
    # Calculate PDF coefficients
    mu_x_2 = np.multiply(mu_x, mu_x)
    mu_x_3 = np.multiply(mu_x_2, mu_x)
    sx_2 = np.multiply(sx, sx)
    sy_2 = np.multiply(sy, sy)
    mu1 = np.divide(np.multiply(mu_y, observed_ratio), mu_x)
    s1 = np.sqrt(
        np.divide(np.multiply(np.multiply(sy_2, observed_ratio), observed_ratio) + sx_2, mu_x_2))
    s1_2 = np.multiply(s1, s1)
    mu1_2 = np.multiply(mu1, mu1)
    A = np.divide(np.multiply(np.multiply(mu_x, observed_ratio), sy_2), mu_x_3)
    B = np.divide(np.multiply(mu_y, sx_2), mu_x_3)
    C = np.sqrt(2./np.pi) * np.divide(
        np.multiply(np.multiply(sx_2, sy),
                    np.exp(-0.5*np.divide(np.multiply(mu_y, mu_y),
                                          sy_2))
                    ),
        np.multiply(mu_x_3, s1_2))

    N = np.multiply(A, mu1) + B + C
    # Calculate mean and standard deviations
    mu = np.divide(np.multiply(A, (s1_2 + mu1_2)) + np.multiply(B, mu1), N)
    s = np.sqrt(np.divide(np.multiply(A, (3*np.multiply(s1_2, mu1)+np.multiply(mu1_2, mu1))) +
                          np.multiply(B, (s1_2 + mu1_2)) +
                          np.multiply(C, np.divide(sx_2, mu_x_2)),
                          N) - np.multiply(mu, mu))
    # Combine mean over stations
    return combine_mu(mu, s)


def combine_mu(mu, s):
    """
    Combine the mean and standard deviations over stations

    Args
        mu: np.array - 2 dimensional numpy array of mean values
        s: np.array - 2 dimensional numpy array of standard deviation values

    Returns
        (np.array, np.array) - tuple of combined mean and standard deviation
            values
    """
    # Loop over each sample
    for i in range(mu.shape[0]):
        if i == 0:
            # First sample so initialise the mean and standard debiation
            combined_mu = mu[i, :]
            combined_s = s[i, :]
        else:
            # Combine with existing values
            si_2 = np.multiply(s[i, :], s[i, :])
            s_2 = np.multiply(combined_s, combined_s)
            combined_mu = (
                np.multiply(combined_mu, si_2)+np.multiply(mu[i, :], s_2)) / (s_2+si_2)
            combined_s = np.multiply(combined_s, s[i, :]) / np.sqrt(s_2+si_2)
    return combined_mu, combined_s


# Supporting PDF functions


def gaussian_pdf(x, mu, sigma):
    """
    Calculate the Gaussian probability

    Calculates the Gaussian probability of x given a mean mu and
    standard deviation sigma.

    Args
        x: Number or np.array of theoretical amplitudes
        mu: Number or list or np.array of means.
            Needs to be same dimensions as x, or a scalar.
        sigma: Number or list or np.array of standard deviations.
            Needs to be same dimensions as x, or a scalar.

    Returns
        float/np.array of probabilities.

    """
    # Avoid nan overflows for 0 errors, make the error very small to get a large number
    if not isinstance(sigma, (float, int, np.float64)):
        sigma[sigma == 0] = _SMALL_NUMBER
    elif sigma == 0:
        sigma = _SMALL_NUMBER
    if isinstance(x, np.ndarray) and x.ndim > 2:
        mu = np.kron(
            np.array([np.kron(mu, np.ones(x.shape[1])).T]).T, np.ones(x.shape[2]))
        sigma = np.kron(
            np.array([np.kron(sigma, np.ones(x.shape[1])).T]).T, np.ones(x.shape[2]))
    return gaussian.pdf(x, loc=mu, scale=sigma)


def gaussian_cdf(x, mu, sigma):
    """
    Calculate the Gaussian Cumulative Distribution Function

    Calculates the Gaussian CDF of x given a mean mu and standard
    deviation sigma.

    Args
        x: Number or np.array of theoretical amplitudes
        mu: Number or list or np.array of means. Needs to be same
                dimensions as x, or a scalar.
        sigma: Number or list or np.array of standard deviations. Needs
                to be same dimensions as x, or a scalar.

    Returns
        float/np.array of CDF values.

    """
    # Avoid nan overflows for 0 errors, make the error very small to get a large number
    if not isinstance(sigma, (float, int, np.float64)):
        sigma[sigma == 0] = _SMALL_NUMBER
    elif sigma == 0:
        sigma = _SMALL_NUMBER
    return gaussian.cdf(x, loc=mu, scale=sigma)


def ratio_pdf(z, mu_x, mu_y, sigma_x, sigma_y, corr=0):
    """
    Calculate the Ratio Probability

    Calculates the Ratio pdf (D. Hinkley, On the ratio of two correlated
    normal random variables, 1969, Biometrika vol 56 pp 635-639).

    Given Z = X/Y and means mu_x, mu_y and standard deviation sigma_x and
    sigma_y. The pdf is normalised.

    Args
        z: Number or numpy array of theoretical amplitudes
        mu_x: Number or list or numpy array of means. Needs to be same
                dimensions as z, or a scalar.
        mu_y: Number or list or numpy array of means. Needs to be same
                dimensions as z, or a scalar.
        sigma_x: Number or list or numpy array of standard deviations.
                Needs to be same dimensions as z, or a scalar.
        sigma_y: Number or list or numpy array of standard deviations.
                Needs to be same dimensions as z, or a scalar.

    Returns
        float/np.array of probabilities

    """
    # Set to prevent numpy warnings
    np.seterr(divide='ignore', invalid='ignore')
    # Expand parameters
    if isinstance(mu_x, np.ndarray) and mu_x.ndim == 3:
        if isinstance(mu_y, np.ndarray) and mu_y.ndim == 2:
            mu_y = np.expand_dims(mu_y, 1)
    if isinstance(mu_y, np.ndarray) and mu_y.ndim == 3:
        if isinstance(mu_x, np.ndarray) and mu_x.ndim == 2:
            mu_x = np.expand_dims(mu_x, 1)
        if isinstance(z, np.ndarray) and z.ndim == 2:
            z = np.expand_dims(z, 1)
        if isinstance(sigma_x, np.ndarray) and sigma_x.ndim == 2:
            sigma_x = np.expand_dims(sigma_x, 1)
        if isinstance(sigma_y, np.ndarray) and sigma_y.ndim == 2:
            sigma_y = np.expand_dims(sigma_y, 1)
    # Calculate some coefficients
    z_2 = np.multiply(z, z)
    sigma_x_2 = np.multiply(sigma_x, sigma_x)
    sigma_xy = np.multiply(sigma_x, sigma_y)
    sigma_y_2 = np.multiply(sigma_y, sigma_y)
    mu_x_2 = np.multiply(mu_x, mu_x)
    if corr > 0:
        a = np.sqrt(np.divide(z_2, sigma_x_2) - 2*corr*np.divide(z, sigma_xy) + 1/sigma_y_2)
    else:
        a = np.sqrt(np.divide(z_2, sigma_x_2) + 1/sigma_y_2)
    a_2 = np.multiply(a, a)
    b = np.divide(np.multiply(mu_x, z), sigma_x_2) - (corr*np.divide(sigma_x_2, sigma_xy)) + \
        np.divide(mu_y, sigma_y_2)
    c = np.divide(mu_x_2, sigma_x_2) + \
        np.divide(np.multiply(mu_y, mu_y), sigma_y_2)
    if corr > 0:
        c -= (2*corr*np.divide(np.multiply(mu_x, mu_y), sigma_xy))
    d = np.exp(np.divide((np.multiply(b, b)-np.multiply(c, a_2)),
                         (2 * (1-corr*corr) * a_2)))
    p = np.divide(np.multiply(b, d),
                  (np.sqrt(2*np.pi) * np.multiply(sigma_xy, np.multiply(a, a_2))))
    p = np.multiply(p,
                    (gaussian_cdf(np.divide(b, (np.sqrt(1-corr*corr) * a)), 0, 1) -
                     gaussian_cdf(np.divide(-b, (np.sqrt(1-corr*corr) * a)), 0, 1))
                    )
    p += np.multiply(np.sqrt(1-corr*corr) / (np.pi*np.multiply(sigma_xy, a_2)),
                     np.exp(-c / (2*(1-corr*corr))))
    if isinstance(p, np.ndarray):
        p[np.isnan(p)] = 0
    # Clear arrays and force a garbage collection to free up memory
    del z_2
    del sigma_x_2
    del sigma_xy
    del sigma_y_2
    del mu_x_2
    del a
    del b
    del c
    del d
    gc.collect()
    return p


def beta_pdf(x, a, b):
    """
    Calculate the Beta Probability

    Calculates the Beta pdf for X given shape parameters a and b.
    Uses scipy.stats.beta.pdf docstring.

    Args
        x: Number or numpy object.
        a: Number or list or numpy object of first shape parameter.
        b: Number or list or numpy object of second shape parameter.

    Returns
        np.array of probabilities

    """
    return beta.pdf(x, a, b)


def heaviside(x):
    """
    Calculate the heaviside step function.

    The heaviside step function is defined as:
        H(x) = 0   x < 0
        H(x) = 0.5 x = 0
        H(x) = 1   x > 0

    Args:
        x: float/np.array - array or float of x values

    Returns:
        float/np.array - array of heaviside values for x values
    """
    return 0.5*(np.sign(x) + 1)


def dkl(ln_probability_p, ln_probability_q, dV=1.0):
    """
    Calculate the Kullback-Liebler divergence for two distributions, p
    and q

    Args
        ln_probability_p: LnPDF - First distribution (p)
        ln_probability_q: LnPDF - Second distribution (q)

    Keyword Args
        dV: float value for the Monte Carlo integration volume element
            (V/N)

    Returns
        float - Kullback-Liebler divergence estimate
    """
    if isinstance(ln_probability_p, LnPDF):
        ln_probability_p = np.ascontiguousarray(
            np.array(ln_probability_p._ln_pdf).flatten())
    if isinstance(ln_probability_q, LnPDF):
        ln_probability_q = np.ascontiguousarray(
            np.array(ln_probability_q._ln_pdf).flatten())
    if cprobability:
        try:
            return cprobability.dkl(ln_probability_p.copy(), ln_probability_q.copy(), dV)
        except Exception:
            logger.exception('WARNING: Error running cython code, resorting to python code')
    else:
        logger.info(C_EXTENSION_FALLBACK_LOG_MSG)
    ind = ln_probability_p > -np.inf
    ln_probability_p = ln_probability_p.copy() - ln_probability_p.max()
    ln_probability_q = ln_probability_q.copy() - ln_probability_q.max()
    probability_p = np.exp(ln_probability_p)
    # Normalise PDF p
    n = np.sum(probability_p)*dV
    ln_probability_p -= np.log(n)
    probability_p /= n
    # Normalise PDF q
    probability_q = np.exp(ln_probability_q)
    n = np.sum(probability_q)*dV
    ln_probability_q -= np.log(n)
    return np.sum(ln_probability_p[ind]*probability_p[ind] -
                  ln_probability_q[ind]*probability_p[ind]) * dV


def ln_marginalise(ln_pdf, axis=0, dV=1.0, _cprob_err=True):
    """
    Marginalise the pdf from the log pdf input

    Marginalises the pdf given a dV, without the dV the normalisation is
    only proportional to the normalised marginalised distribution.

    Args
        pdf: np.array/LnPDF - array or object that can be multiplied by dV


    Optional Args
        axis: integer - axis over which to marginalise [default = 0]
        dV: float - volume element in marginalisation, default value is
                 1.0 giving a proportional marginalised distribution.

    Returns
        np.array - marginalised distribtion over axis
    """
    if cprobability and axis == 0:
        try:
            if isinstance(ln_pdf, LnPDF):
                return cprobability.ln_marginalise(ln_pdf._ln_pdf.astype(np.float64))
            return cprobability.ln_marginalise(ln_pdf.astype(np.float64))
        except Exception:
            if _cprob_err:
                logger.exception('WARNING: Error running cython code, resorting to python code')
    else:
        logger.info(C_EXTENSION_FALLBACK_LOG_MSG)
    # scale and then marginalise:
    ln_scale = 0
    if ln_pdf.shape[axis] == 1:
        return np.array(ln_pdf).squeeze(axis)
    elif axis == 0 and len(ln_pdf.shape) == 1:
        # Probably shouldn't be marginalising over the last parameter
        return ln_pdf
    if np.prod(ln_pdf.shape) and ln_pdf.max() < 0 and ln_pdf.max() > -np.inf:
        ln_scale = -ln_pdf.max()
    with warnings.catch_warnings() and np.errstate(divide='ignore'):
        warnings.simplefilter("ignore")
        # if axis == 0:
        #     # Get consistency with c code
        #     ln_pdf = np.array(ln_pdf)
        result = np.log(np.sum(np.exp(ln_pdf+ln_scale)*dV, axis=axis)) - ln_scale
    if axis == 0:
        result = np.array(result).flatten()
    return result


def ln_normalise(ln_pdf, dV=1):
    """
    Normalise the pdf the pdf from the log pdf input.

    Normalises the pdf given a dV without the dV the normalisation is
    only proportional to the normalised distribution.

    Args
        pdf: Object that can be multiplied by dV (numpy array/matrix or PDF object)

    Optional Args
        dV: float - volume element in normalisation, default value is
                 1.0 giving a proportional normalised distribution.
    Returns
        np.array - normalised distribtion

    """
    if ln_pdf.ndim == 1 and ln_pdf.shape[0] == 1:
        return np.array([0])
    if cprobability:
        try:
            if ln_pdf.ndim != 1 and ln_pdf.shape[0] != 1:
                raise Exception('Incorrect shape')
            if ln_pdf.ndim == 2:
                normalised_ln_pdf = cprobability.ln_normalise(np.asarray(ln_pdf).flatten())
            else:
                normalised_ln_pdf = cprobability.ln_normalise(ln_pdf)
            return normalised_ln_pdf
        except Exception:
            logger.exception('WARNING: Error running cython code, resorting to python code')
    else:
        logger.info(C_EXTENSION_FALLBACK_LOG_MSG)
    # scale and then marginalise:
    ln_scale = 0
    if -ln_pdf.max() > 0 and ln_pdf.max() > -np.inf:
        ln_scale = -ln_pdf.max()
    with warnings.catch_warnings() and np.errstate(divide='ignore'):
        warnings.simplefilter("ignore")
        ln_n = np.log(np.sum(np.exp(ln_pdf+ln_scale) * dV, axis=None)) - ln_scale
    if ln_pdf.ndim == 2 and 1 in list(ln_pdf.shape):
        ln_pdf = np.squeeze(np.asarray(ln_pdf))
        if not ln_pdf.shape:
            ln_pdf = np.array([ln_pdf])
    # ln_probability_scale_factor is automatically included in normalisation
    return ln_pdf - ln_n


def model_probabilities(*args):
    """
    Calculate the model probabilities for a discrete set of models using the
    ln_bayesian_evidences, provided as args.

    e.g. to compare between DC and MT:

        pDC,pMT=model_probabilities(dc_ln_bayesian_evidence,mt_ln_bayesian_evidence)

    Args
        floats - ln_bayesian_evidence for each model type

    Returns
        tuple: Tuple of the normalised model probabilities for the corresponding
            ln_bayesian_evidence inputs
    """
    max_LnBE = -np.inf
    output_model_probabilities = []
    for ln_bayesian_evidence in args:
        max_LnBE = max([max_LnBE, ln_bayesian_evidence])
    norm = 0
    for ln_bayesian_evidence in args:
        p = np.exp(ln_bayesian_evidence-max_LnBE)
        output_model_probabilities.append(p)
        norm += p
    for i, model_probability in enumerate(output_model_probabilities):
        output_model_probabilities[i] = model_probability/norm
    return tuple(output_model_probabilities)


def dkl_estimate(ln_pdf, V, N):
    """
    Calculate the Kullback-Leibeler divergence for the ln_pdf from the
    prior distribution.

    Assumes that the moment tensor prior is uniform in the sampling
    and that is the form that the PDFs have been evaluated in.

    Args
        ln_pdf: input LnPDF object
        V: float value for the Volume of the sample space for Monte
            Carlo delta V and prior estimates
        N: integer number of tried samples

    """
    dV = V/float(N)
    # Doesnt use dkl function as can be simplified to reduce calculation
    if isinstance(ln_pdf, LnPDF):
        ln_pdf = np.ascontiguousarray(np.array(ln_pdf._ln_pdf).flatten())
    if cprobability:
        try:
            return cprobability.dkl_uniform(ln_pdf.copy(), V, dV)
        except Exception:
            logger.exception('WARNING: Error running cython code, resorting to python code')
    else:
        logger.info(C_EXTENSION_FALLBACK_LOG_MSG)
    ind = ln_pdf > -np.inf
    ln_pdf = ln_pdf[ind].copy()-ln_pdf.max()
    pdf = np.exp(ln_pdf)
    # Normalise PDF
    n = np.sum(pdf)*dV
    ln_pdf -= np.log(n)
    pdf /= n
    return np.sum(np.array(ln_pdf)*np.array(pdf) + np.array(pdf)*np.log(V), -1)*dV


class LnPDF(object):

    """
    LnPDF object, to allow arithmetic operations on the pdf

    A simple object containing a pdf and several methods for acting on the pdf.
    _ln_pdf is a numpy matrix containing the natural logarithm of the pdf
    """

    def __init__(self, ln_pdf=None, pdf=None, dV=1, *args, **kwargs):
        """LnPDF Initialisation

        Initialises the LnPDF object.

        Optional Args
            ln_pdf:[None] numpy array/matrix or list containing log pdf samples
            pdf:[None] numpy array/matrix or list containing pdf samples
            dV:[1] default volume element for normalisation
        """
        # Possibly good to add initialisation size arg for LnPDF,
        # initSize=(1,1)
        self._ln_pdf = np.matrix([])
        if ln_pdf is not None:
            self._set_ln_pdf(ln_pdf)
        elif pdf is not None:
            with warnings.catch_warnings() and np.errstate(divide='ignore'):
                warnings.simplefilter("ignore")
                self._set_ln_pdf(np.log(pdf))
        if dV:
            self._set_dv(dV)
        else:
            self._set_dv(1)

    def __getstate__(self):
        return self._ln_pdf, self.dV

    def __setstate__(self, ln_pdf, dV):
        self._set_ln_pdf(ln_pdf)
        self._set_dv(dV)

    def __getattr__(self, key):
        """x.__getattr__(y) <==> x.y"""
        # Handles cases not picked up by __getattribute__
        if key == 'shape':
            return self._ln_pdf.shape

    def __len__(self):
        """
        Return the length of ln_pdf

        Returns
            int - length of ln_pdf (axis=-1)

        """
        if isinstance(self._ln_pdf, np.ndarray):
            return self.shape[-1]
        return 1

    def __repr__(self):
        """x.__repr__() <==> repr(x)"""
        return self._ln_pdf.__repr__()

    def __getitem__(self, index):
        """x.__getitem__(index) <==> x[index]"""
        return self._ln_pdf.__getitem__(index)

    def __getslice__(self, i, j):
        """x.__getslice__(i, j) <==> x[i:j]"""
        return self._ln_pdf.__getslice__(i, j)

    def __setitem__(self, index, val):
        """x.__setitem__(index, val) <==> x[index] = val"""
        return self._ln_pdf.__setitem__(index, val)

    def __setslice__(self, i, j, val):
        """x.__setslice__(i, j, val) <==> x[i, j] = val"""
        return self._ln_pdf.__setslice__(i, j, val)

    def _cmp(self, other, operator_func):
        """
        Base comparison function.

        Args
            other: object - object for comparison
            operator_func: func - function from the operator module to compare using

        Returns:
            boolean: result of the comparison
        """
        if isinstance(other, self.__class__):
            return operator_func(self._ln_pdf, other._ln_pdf)
        else:
            return operator_func(self._ln_pdf, other)

    def __lt__(self, other):
        """x.__lt__(y) <==> x < y"""
        return self._cmp(other, operator.__lt__)

    def __gt__(self, other):
        """x.__gt__(y) <==> x > y"""
        return self._cmp(other, operator.__gt__)

    def __le__(self, other):
        """x.__le__(y) <==> x <= y"""
        return self._cmp(other, operator.__le__)

    def __ge__(self, other):
        """x.__ge__(y) <==> x >= y"""
        return self._cmp(other, operator.__ge__)

    def __eq__(self, other):
        """x.__eq__(y) <==> x == y"""
        return self._cmp(other, operator.__eq__)

    def __ne__(self, other):
        """x.__ne__(y) <==> x != y"""
        return self._cmp(other, operator.__ne__)

    def _arithmetic(self, other, arithmetic_func):
        """
        Base comparison function.

        Args
            other: object - object for comparison
            operator_func: func - function from the operator module to compare using

        Returns:
            boolean: result of the comparison
        """
        if isinstance(other, self.__class__):
            if self.dV == other.dV:
                return self.__class__(ln_pdf=arithmetic_func(self._ln_pdf, other._ln_pdf), dV=self.dV)
            else:
                raise ValueError(
                    'Incorrect dV, should match between self and other')
        else:
            return self.__class__(ln_pdf=arithmetic_func(self._ln_pdf, other), dV=self.dV)

    def __mul__(self, other):
        """x.__mul__(y) <==> x*y"""
        return self._arithmetic(other, np.multiply)

    def __div__(self, other):
        """x.__div__(y) <==> x/y"""
        return self._arithmetic(other, np.divide)

    def __truediv__(self, other):
        """x.__div__(y) <==> x/y"""
        return self.__div__(other)

    def __add__(self, other):
        """x.__add__(y) <==> x+y"""
        return self._arithmetic(other, operator.__add__)

    def __sub__(self, other):
        """x.__sub__(y) <==> x-y"""
        return self._arithmetic(other, operator.__sub__)

    def __rsub__(self, other):
        """x.__rsub__(y) <==> y-x"""
        return (self-other)*-1

    def __radd__(self, other):
        """x.__radd__(y) <==> y+x"""
        return (self+other)

    def __rmul__(self, other):
        """x.__rmul__(y) <==> y*x"""
        return self*other

    def __rdiv__(self, other):
        """x.__rdiv__(2) <==> 2/x"""
        return LnPDF(ln_pdf=(other/self._ln_pdf), dV=self.dV)

    def __rtruediv__(self, other):
        """x.__rdiv__(2) <==> 2/x"""
        return self.__rdiv__(other)

    def __abs__(self):
        """x.__abs__() <==> abs(x)"""
        return LnPDF(ln_pdf=np.abs(self._ln_pdf), dV=self.dV)

    def __float__(self):
        """
        Convert pdf to float

        Returns
            Converted LnPDF to float (if a unit object, else error)

        Raises
            TypeError: array not length 1 so cannot be converted to a scalar.

        """
        return float(self._ln_pdf)

    def argmax(self, axis=1):
        """
        Return the index of maximum value over an axis

        Returns the indices of the maximum values from the pdf over the
        given axis. If the axis is set to -1 returns the index of the
        maximum value over the marginalised pdf.

        Optional Args
            axis: int - axis over which to find the argmax [default=1]

        Returns
            np.array - array of indices to maximum values.
        """
        axis = min(self._ln_pdf.ndim-1, axis)
        if self._ln_pdf.ndim > 1 and axis == -1:
            return self.marginalise().argmax(0).flatten()
        return self._ln_pdf.argmax(axis)

    def max(self, axis=1):
        """Return the maximum values over a given axis

        Returns the maximum values from the LnPDF over the given axis.
        If the axis is set to -1 returns the maximum value over the
        marginalised pdf.

        Optional Args
            axis: int - axis over which to find the argmax [default=1]

        Returns
            np.array - array of maximum values.

        """
        if not np.prod(self._ln_pdf.shape):
            return 0
        axis = min(self._ln_pdf.ndim-1, axis)
        if self._ln_pdf.ndim > 1 and axis == -1:
            return np.exp(self.marginalise().max()).flatten()
        return float(np.exp(self._ln_pdf.max(axis)))

    def _set_dv(self, dV):
        """Private Function to set dV value"""
        self.dV = dV

    def _set_ln_pdf(self, ln_pdf):
        """Private Function to set _ln_pdf value"""
        if isinstance(ln_pdf, self.__class__):
            ln_pdf = ln_pdf._ln_pdf
        self._ln_pdf = np.matrix(ln_pdf)

    def output(self, normalise=True):
        """
        Return the marginalised pdf for outputting.

        Optional Args
            normalise: bool - flag to normalise the PDF before
                    outputting [default = True]
        """
        if normalise:
            return self.marginalise().normalise()
        return self.marginalise()

    def exp(self, _cprob_err=True):
        if cprobability:
            try:
                return cprobability.ln_exp(self._ln_pdf)
            except Exception:
                if _cprob_err:
                    logger.exception('WARNING: Error running cython code, resorting to python code')
        else:
            logger.info(C_EXTENSION_FALLBACK_LOG_MSG)
        return np.exp(self._ln_pdf)

    def nonzero(self, discard=100000., n_samples=0, _cprob_err=True):
        """
        Return the non-zero indices of the pdf

        Returns the non zero indices of the marginalised pdf.

        Samples with probability less than n_samples*discard of the max
        probability can be discarded.

        Optional Args
            discard: float - discard scale [default = 100000.]
            n_samples: integer - number of samples generated [default = 0]
        """
        ln_pdf = np.array(self.marginalise(axis=0, _cprob_err=_cprob_err)._ln_pdf).flatten()
        m_val = -np.inf
        if n_samples > 0 and discard > 0:
            m_val = max(ln_pdf) - np.log(discard*n_samples)
        return np.nonzero(1*(ln_pdf > m_val))[0]

    def normalise(self, dV=False):
        """
        Normalise the pdf object

        Optional Args
            dV: float - update the object's dV value.

        Returns
            LnPDF object with normalised pdf.
        """
        if dV:
            self._set_dv(dV)
        new = self.__class__(dV=self.dV)
        new._ln_pdf = ln_normalise(self._ln_pdf, self.dV)
        return new

    def marginalise(self, axis=0, dV=False, _cprob_err=True):
        """
        Marginalise the pdf object over a given axis

        Optional Args
            axis: integer - axis over which to marginalise [default = 0]
            dV: float - update the object's dV value.

        Returns
            LnPDF object with marginalised pdf.
        """
        if dV:
            self._set_dv(dV)
        new = self.__class__(dV=self.dV)
        new._ln_pdf = ln_marginalise(self._ln_pdf, axis=axis, dV=self.dV, _cprob_err=_cprob_err)
        return new

    def append(self, other, axis=1):
        """
        Append values to pdf

        Args
            other: LnPDF/np.array - Object to append to ln_pdf

        Optional Args
            axis: integer - axis over which to append [default = 1]
        """
        if not self.shape[1]:
            if isinstance(other, self.__class__):
                self._ln_pdf = other._ln_pdf
                self.dV = other.dV
            elif isinstance(other, np.ndarray):
                self._ln_pdf = other
        else:
            if isinstance(other, self.__class__) and self.dV == other.dV:
                self._ln_pdf = np.append(
                    self._ln_pdf, other._ln_pdf, axis=axis)
            elif isinstance(other, np.ndarray):
                self._ln_pdf = np.append(self._ln_pdf, other, axis=axis)
